#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
ISO STEP reader/writer.
"""

import time
from typing import List
from collections import deque
from dataclasses import dataclass, field

import matplotlib.pyplot as plt
import networkx as nx
import plot_data.graph

import dessia_common.core as dc  # isort: skip
from dessia_common.files import BinaryFile  # isort: skip

import volmdlr
import volmdlr.core
import volmdlr.edges
import volmdlr.faces
import volmdlr.curves
import volmdlr.primitives3d
import volmdlr.wires
from volmdlr.utils import step_reader
from volmdlr.utils.step_reader import STEP_TO_VOLMDLR, STEP_REPRESENTATION_ENTITIES


class StepFunction(dc.DessiaObject):
    """
    Abstract class defining a step function.

    """

    def __init__(self, function_id, function_name, function_arg):
        self.id = function_id
        self.name = function_name
        self.arg = function_arg

        # TODO : modify this continuation and simplify
        if self.name == "":
            if self.arg[1][0] == 'B_SPLINE_SURFACE':
                self.simplify('B_SPLINE_SURFACE')
            if self.arg[1][0] == 'B_SPLINE_CURVE':
                self.simplify('B_SPLINE_CURVE')
        dc.DessiaObject.__init__(self, name=function_name)

    def simplify(self, new_name):
        """ADD DOCSTRING."""
        # ITERATE ON SUBFUNCTIONS
        args = [subfun[1] for (i, subfun) in enumerate(self.arg) if
                (len(subfun[1]) != 0 or i == 0)]
        arguments = []
        for arg in args:
            if not arg:
                arguments.append("''")
            else:
                arguments.extend(arg)
        arguments.pop()  # DELETE REPRESENTATION_ITEM('')

        self.name = new_name
        self.arg = arguments


class Step(dc.DessiaObject):
    """
    Defines the Step class.

    """

    _standalone_in_db = True

    def __init__(self, lines: List[str], name: str = ''):
        self.functions, self.connections = self.read_lines(lines)
        self._graph = None
        self.global_uncertainty = 1e-6
        self.length_conversion_factor = 1
        self.angle_conversion_factor = 1
        # self.read_diagnostic = StepReaderReport
        self._roots_nodes = None

        dc.DessiaObject.__init__(self, name=name)

    @property
    def all_connections(self):
        """Returns all pairs of connections."""
        list_connections = []
        for key, values in self.connections.items():
            for value in values:
                list_connections.append((key, value))
        return list_connections

    @property
    def root_nodes(self):
        """Returns a dictionary containing the nodes of the step file function that are used as start points."""
        if not self._roots_nodes:
            self._roots_nodes = self.get_root_nodes()
        return self._roots_nodes

    def graph(self):
        """Returns the step file networkx graph of dependencies."""
        if not self._graph:
            self._graph = self.create_graph()
        return self._graph

    @classmethod
    def from_stream(cls, stream: BinaryFile, name: str = ''):
        """Instantiate a Step object from a stream."""
        stream.seek(0)
        lines = []
        for line in stream:
            line = line.decode("ISO-8859-1")
            line = line.replace("\r", "")
            lines.append(line)
        return cls(lines, name=name)

    @classmethod
    def from_file(cls, filepath: str = None, name: str = ''):
        """Instantiate a Step object from a step file."""
        with open(filepath, "r", encoding="ISO-8859-1") as file:
            lines = []
            for line in file:
                lines.append(line)
        return cls(lines, name=name)

    def read_lines(self, lines):
        """Translate the step file into step functions objects."""
        dict_connections = {}
        previous_line = ""
        functions = {}
        for line in lines:
            # line = line.replace(" ", "")
            line = line.replace("\n", "")

            # SKIP EMPTY LINE
            if not line:
                continue

            # ASSEMBLE LINES IF THEY ARE SEPARATED
            if line[-1] != ';':
                previous_line = previous_line + line
                continue

            line = previous_line + line

            # SKIP HEADER
            if line[0] != "#":
                previous_line = str()
                continue

            function = line.split("=", maxsplit=1)
            function_id = int(function[0][1:].strip())
            function_name_arg = function[1].split("(", 1)
            function_name = function_name_arg[0].replace(" ", "")
            start_index_name = function_name_arg[1].find("'")
            if start_index_name != -1:
                end_index_name = function_name_arg[1].find("'", start_index_name + 1)
                if end_index_name != -1:
                    function_arg_string = function_name_arg[1][end_index_name + 1:]
                else:
                    function_arg_string = function_name_arg[1]
            else:
                function_arg_string = function_name_arg[1]
            function_arg = function_arg_string.split("#")
            connections = []
            for connec in function_arg[1:]:
                connec = connec.split(",")
                connec = connec[0].split(")")
                if connec[0][-1] != "'":
                    function_connection = int(connec[0])
                    connections.append(function_connection)

            previous_line = str()

            # FUNCTION ARGUMENTS
            functions, connections = self._helper_intantiate_step_functions(functions, connections,
                                                                            [function_id, function_name,
                                                                             function_name_arg])

            dict_connections[function_id] = connections

        return functions, dict_connections

    def _helper_intantiate_step_functions(self, functions, connections, function_parameters):
        """Helper function to read_lines."""
        function_id, function_name, function_name_arg = function_parameters
        function_arg = function_name_arg[1]
        arguments = step_reader.step_split_arguments(function_arg)
        new_name = ''
        new_arguments = []
        if function_name == "":
            name_arg = self.step_subfunctions(arguments)
            for name, arg in name_arg:
                new_name += name + ', '
                new_arguments.extend(arg)
            new_name = new_name[:-2]
            function_name = new_name
            arguments = new_arguments
            for arg in arguments:
                if arg[0] == '#':
                    connections.append(int(arg[1:]))

        for i, argument in enumerate(arguments):
            if argument[:2] == '(#' and argument[-1] == ')':
                arg_list = step_reader.set_to_list(argument)
                arguments[i] = arg_list

        function = StepFunction(function_id, function_name, arguments)
        functions[function_id] = function
        return functions, connections

    def not_implemented(self):
        not_implemented = []
        for _, fun in self.functions.items():
            if fun.name not in STEP_TO_VOLMDLR:
                not_implemented.append(fun.name)
        return list(set(not_implemented))

    def create_graph(self):
        """
        Step functions graph.

        :return: A graph representation the step file structure.
        :rtype: networkx.DiGraph
        """
        graph = nx.DiGraph()
        labels = {}

        for function in self.functions.values():
            if function.name == 'SHAPE_REPRESENTATION_RELATIONSHIP':
                # Create short cut from id1 to id2
                id1 = int(function.arg[2][1:])
                id2 = int(function.arg[3][1:])
                elem1 = (function.id, id1)
                elem2 = (function.id, id2)
                self.all_connections.remove(elem1)
                self.all_connections.remove(elem2)
                self.all_connections.append((elem1[1], elem2[1]))

                self.functions[id1].arg.append(f'#{id2}')

            elif function.name in STEP_TO_VOLMDLR:
                graph.add_node(function.id,
                               color='rgb(0, 0, 0)',
                               shape='.',
                               name=str(function.id))
                labels[function.id] = str(function.id) + ' ' + function.name

        # Delete connection if node not found
        node_list = list(graph.nodes())
        delete_connection = []
        for connection in self.all_connections:
            if connection[0] not in node_list \
                    or connection[1] not in node_list:
                delete_connection.append(connection)
        for delete in delete_connection:
            self.all_connections.remove(delete)

        # Create graph connections
        graph.add_edges_from(self.all_connections)

        # Remove single nodes
        delete_nodes = []
        for node in graph.nodes:
            if graph.degree(node) == 0:
                delete_nodes.append(node)
        for node in delete_nodes:
            graph.remove_node(node)
            # G.remove_node(node)
        return graph

    def draw_graph(self, graph=None, reduced=False):
        """
        Draw a graph for Step data.

        :param graph: DESCRIPTION, defaults to None
        :type graph: TYPE, optional
        :param reduced: DESCRIPTION, defaults to False
        :type reduced: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """

        delete = ['CARTESIAN_POINT', 'DIRECTION']
        if graph is None:
            new_graph = self.create_graph()
        else:
            new_graph = graph.copy()

        labels = {}
        for id_nb, function in self.functions.items():
            if id_nb in new_graph.nodes and not reduced:
                labels[id_nb] = str(id_nb) + ' ' + function.name
            elif id_nb in new_graph.nodes and reduced:
                if function.name not in delete:
                    labels[id_nb] = str(id_nb) + ' ' + function.name
                else:
                    new_graph.remove_node(id_nb)
        pos = nx.kamada_kawai_layout(new_graph)
        plt.figure()
        nx.draw_networkx_nodes(new_graph, pos)
        nx.draw_networkx_edges(new_graph, pos)
        nx.draw_networkx_labels(new_graph, pos, labels)

    @staticmethod
    def step_subfunctions(subfunctions):
        """Handles context elements from step file."""
        subfunctions = subfunctions[0].replace(" ", "")
        parenthesis_count = 0
        subfunction_names = []
        subfunction_args = []
        subfunction_name = ""
        subfunction_arg = ""
        for char in subfunctions:

            if char == "(":
                parenthesis_count += 1
                if parenthesis_count == 1:
                    subfunction_names.append(subfunction_name)
                    subfunction_name = ""
                else:
                    subfunction_arg += char

            elif char == ")":
                parenthesis_count -= 1
                if parenthesis_count == 0:
                    subfunction_args.append(subfunction_arg)
                    subfunction_arg = ""
                else:
                    subfunction_arg += char

            elif parenthesis_count == 0:
                subfunction_name += char

            else:
                subfunction_arg += char
        return [
            (subfunction_names[i], step_reader.step_split_arguments(subfunction_args[i]))
            for i in range(len(subfunction_names))]

    def parse_arguments(self, arguments):
        """Converts the arguments IDs from string to integer."""
        for i, arg in enumerate(arguments):
            if isinstance(arg, str) and arg[0] == '#':
                arguments[i] = int(arg[1:])
            elif isinstance(arg, str) and arg[0:2] == '(#':
                argument = []
                arg_id = ""
                for char in arg[1:-1]:
                    if char == ',':
                        argument.append(arg_id)
                        arg_id = ""
                        continue

                    arg_id += char
                argument.append(arg_id)
                arguments[i] = argument

    def instantiate(self, name, arguments, object_dict, step_id):
        """
        Gives the volmdlr object related to the step function.
        """
        self.parse_arguments(arguments)
        fun_name = name.replace(', ', '_')
        fun_name = fun_name.lower()
        try:
            if hasattr(step_reader, fun_name):
                volmdlr_object = getattr(step_reader, fun_name)(arguments, object_dict)

            elif name in STEP_TO_VOLMDLR and hasattr(STEP_TO_VOLMDLR[name], "from_step"):
                volmdlr_object = STEP_TO_VOLMDLR[name].from_step(
                    arguments, object_dict, name=name, step_id=step_id, global_uncertainty=self.global_uncertainty,
                    length_conversion_factor=self.length_conversion_factor,
                    angle_conversion_factor=self.angle_conversion_factor)

            else:
                raise NotImplementedError(f'Dont know how to interpret #{step_id} = {name}({arguments})')
        except (ValueError, NotImplementedError) as error:
            raise ValueError(f"Error while instantiating #{step_id} = {name}({arguments})") from error
        return volmdlr_object

    def create_node_list(self, initial_nodes):
        """
        Step functions graph as a list of nodes.

        :param initial_nodes: Initial list of shell nodes and assemblies entities.
        :type initial_nodes: List[int]
        :return: A list of nodes in the right order of dependency.
        :rtype: List[int]
        """
        list_head = []
        list_nodes = []
        visited_set = set()
        stack = deque(initial_nodes)
        while stack:
            node = stack.popleft()
            name = self.functions[node].name
            if node not in visited_set and name in STEP_TO_VOLMDLR:
                visited_set.add(node)
                if self.connections[node]:
                    list_nodes.append(node)
                    for connection in self.connections[node]:
                        if connection not in visited_set:
                            stack.append(connection)
                else:
                    # Entities without connections should be instantiated first
                    list_head.append(node)
        return list_head + list_nodes[::-1]

    def get_shell_node_from_representation_entity(self, id_representation_entity: int):
        """
        Find the shell node ID related to a given representation entity.

        :param id_representation_entity: Representation entity ID.
        :type id_representation_entity: int
        :return: Shell ID.
        :rtype: int
        """
        name_representation_entity = self.functions[id_representation_entity].name
        arg = self.functions[id_representation_entity].arg[1]
        if name_representation_entity == "MANIFOLD_SURFACE_SHAPE_REPRESENTATION":
            if self.functions[int(arg[0][1:])].name == "AXIS2_PLACEMENT_3D":
                id_solid_entity = int(arg[1][1:])
            else:
                id_solid_entity = int(arg[0][1:])
            if self.functions[id_solid_entity].name in {"CLOSED_SHELL", "OPEN_SHELL"}:
                return id_solid_entity
            id_shell = self.functions[id_solid_entity].arg[1]
            if isinstance(id_shell, list):
                return int(id_shell[0][1:])
            return int(id_shell[1:])
        if self.functions[int(arg[0][1:])].name == "AXIS2_PLACEMENT_3D":
            id_solid_entity = int(arg[1][1:])
        else:
            id_solid_entity = int(arg[0][1:])
        id_shell = self.functions[id_solid_entity].arg[1]
        if isinstance(id_shell, list):
            return int(id_shell[0][1:])
        return int(id_shell[1:])

    def get_shell_node_from_shape_representation(self, id_shape_representation: int):
        """
        Find the shell node ID related to a given shape representation.

        :param id_shape_representation: Representation entity ID.
        :type id_shape_representation: int
        :return: Shell ID.
        :rtype: int
        """
        if len(self.functions[id_shape_representation].arg) < 4:
            # From the step file, the SHAPE_REPRESENTATION entity has 3 arguments. But we add a 4th argument to
            # those SHAPE_REPRESENTATION entity that are related to a representation entity. So, if the length of arg
            # is less of 4 there is no representation entity related to it, and we return None.
            return None
        id_representation_entity = int(self.functions[id_shape_representation].arg[3][1:])
        id_solid_entity = int(self.functions[id_representation_entity].arg[1][0][1:])
        id_shell = self.functions[id_solid_entity].arg[1]
        if isinstance(id_shell, list):
            return int(id_shell[0][1:])
        return int(id_shell[1:])

    def get_frame_mapped_shell_node(self, node: int):
        """
        Find the shell node in the assembly.

        :param node: Assembly step entity node.
        :type node: int
        :return: Shell ID.
        :rtype: int
        """
        id_representation_entity = None
        arguments = self.functions[node].arg
        name_arg1 = self.functions[int(arguments[2][1:])].name
        name_arg2 = self.functions[int(arguments[3][1:])].name
        if name_arg1 in STEP_REPRESENTATION_ENTITIES:
            id_representation_entity = int(arguments[2][1:])
        elif name_arg2 in STEP_REPRESENTATION_ENTITIES:
            id_representation_entity = int(arguments[3][1:])
        if id_representation_entity:
            return self.get_shell_node_from_representation_entity(id_representation_entity)
        id_shape_representation = int(arguments[3][1:])
        if len(self.functions[id_shape_representation].arg) < 4:
            id_shape_representation = int(arguments[2][1:])
        if self.functions[id_shape_representation].name == "SHAPE_REPRESENTATION":
            return self.get_shell_node_from_shape_representation(id_shape_representation)
        id_representation_entity = int(self.functions[id_shape_representation].arg[1][1][1:])
        id_shell = self.functions[id_representation_entity].arg[1]
        if isinstance(id_shell, list):
            return int(id_shell[0][1:])
        return int(id_shell[1:])

    def shape_definition_representation_to_shell_node(self, shape_definition_representation_id):
        """Returns the ID of the shell entity related to the given shape_definition_representation ID."""
        id_representation_entity = self.functions[shape_definition_representation_id].arg[1]
        function_name = self.functions[int(id_representation_entity[1:])].name
        if function_name in STEP_REPRESENTATION_ENTITIES:
            return self.get_shell_node_from_representation_entity(int(id_representation_entity[1:]))
        if function_name == "SHAPE_REPRESENTATION":
            return self.get_shell_node_from_shape_representation(int(id_representation_entity[1:]))
        raise NotImplementedError("This shape representation has not yet been recognized by volmdlr step translator.")

    def product_definition_to_product(self, id_product_definition):
        """Returns the ID of the product entity related to the given product_definition ID."""
        if self.functions[id_product_definition].name == "NEXT_ASSEMBLY_USAGE_OCCURRENCE":
            id_product_definition = int(self.functions[id_product_definition].arg[3][1:])
        id_product_definition_formation = self.functions[id_product_definition].arg[2]
        id_product = self.functions[int(id_product_definition_formation[1:])].arg[2]
        return int(id_product[1:])

    def shape_definition_representation_to_product_node(self, shape_definition_representation_id):
        """Returns the ID of the product entity related to the given shape_definition_representation ID."""
        id_product_definition_shape = self.functions[shape_definition_representation_id].arg[0]
        id_product_definition = int(self.functions[int(id_product_definition_shape[1:])].arg[2][1:])
        return self.product_definition_to_product(id_product_definition)

    def get_root_nodes(self):
        """Returns a dictionary containing the nodes of the step file function that are used as start points."""
        next_assembly_usage_occurrence = []
        product_definitions = []
        shape_representation_relationship = []
        shape_representations = []
        shape_definition_representation = []
        shell_nodes = []
        geometric_representation_context = {}
        not_shell_nodes = []
        context_dependent_shape_representation = []
        for function in self.functions.values():
            if function.name == "NEXT_ASSEMBLY_USAGE_OCCURRENCE":
                next_assembly_usage_occurrence.append(function.id)
            elif function.name == "PRODUCT_DEFINITION":
                product_definitions.append(function.id)
            elif function.name == "SHAPE_REPRESENTATION_RELATIONSHIP":
                shape_representation_relationship.append(function.id)
                id_shape_representation = int(function.arg[3][1:])
                shape_representations.append(id_shape_representation)
            elif function.name == "SHAPE_DEFINITION_REPRESENTATION":
                id_shape_representation = int(function.arg[1][1:])
                shape_representations.append(id_shape_representation)
                id_geometric_context = int(self.functions[id_shape_representation].arg[-1][1:])
                geometric_representation_context[id_shape_representation] = id_geometric_context
                shape_definition_representation.append(function.id)
            elif function.name in {"CLOSED_SHELL", "OPEN_SHELL"}:
                shell_nodes.append(function.id)
            elif function.name == 'BREP_WITH_VOIDS':
                shell_nodes.append(function.id)
                not_shell_nodes.append(int(function.arg[1][1:]))
            elif function.name == "CONTEXT_DEPENDENT_SHAPE_REPRESENTATION":
                context_dependent_shape_representation.append(function.id)
        for node in not_shell_nodes:
            shell_nodes.remove(node)
        return {"NEXT_ASSEMBLY_USAGE_OCCURRENCE": next_assembly_usage_occurrence,
                "CONTEXT_DEPENDENT_SHAPE_REPRESENTATION": context_dependent_shape_representation,
                "PRODUCT_DEFINITION": product_definitions,
                "SHAPE_REPRESENTATION_RELATIONSHIP": shape_representation_relationship,
                "SHAPE_REPRESENTATION": shape_representations,
                "SHAPE_DEFINITION_REPRESENTATION": shape_definition_representation,
                "GEOMETRIC_REPRESENTATION_CONTEXT": geometric_representation_context,
                "SHELLS": shell_nodes}

    def get_assembly_structure(self):
        """
        Get assembly dependency structure.
        """
        assemblies_structure = {}
        assemblies = set()
        shapes = set()
        for node in self.root_nodes["NEXT_ASSEMBLY_USAGE_OCCURRENCE"]:
            function = self.functions[node]
            assembly_product_definition = int(function.arg[3][1:])
            assembly_node = int(self.functions[assembly_product_definition].arg[4][1:])
            assemblies_structure.setdefault(assembly_node, []).append(node)
            assemblies.add(assembly_node)
            id_product_definition = int(function.arg[4][1:])
            if len(self.functions[id_product_definition].arg) > 5:
                for arg in self.functions[id_product_definition].arg[5:]:
                    shapes.add(int(arg[1:]))
        valid_entities = assemblies.union(shapes)
        return assemblies_structure, valid_entities

    def get_assembly_data(self, assembly_usage_occurence, valid_entities, assembly_frame, object_dict):
        """
        Helper function to get assembly data.
        """
        assembly_shapes = []
        assembly_positions = []
        for node in assembly_usage_occurence:
            function = self.functions[node]
            id_product_definition = int(function.arg[4][1:])
            ids_shape_definition_representation = [int(arg[1:]) for
                                                   arg in self.functions[id_product_definition].arg[4:]
                                                   if int(arg[1:]) in valid_entities]
            assembly_shapes.extend(ids_shape_definition_representation)
            id_context_dependent_shape_representation = int(function.arg[-1][1:])
            id_transformation = int(self.functions[id_context_dependent_shape_representation].arg[0][1:])
            id_item_defined_transformation = int(self.functions[id_transformation].arg[4][1:])
            id_frame1 = int(self.functions[id_item_defined_transformation].arg[2][1:])
            id_frame2 = int(self.functions[id_item_defined_transformation].arg[3][1:])
            frame1 = object_dict[id_frame1]
            # frame2 = object_dict[id_frame2]
            if frame1 == assembly_frame:
                component_frame = id_frame2
            else:
                component_frame = id_frame1
            positions = [component_frame] * len(ids_shape_definition_representation)
            assembly_positions.extend(positions)
        return assembly_shapes, assembly_positions

    def context_dependent_shape_representation_to_next_assembly_usage_occurrence(self, node):
        """
        Returns id of the next_assembly_usage_occurrence related to the given context_dependent_shape_representation.
        """
        arg = self.functions[node].arg
        id_product_definition_shape = int(arg[1][1:])
        return int(self.functions[id_product_definition_shape].arg[2][1:])

    def create_connections(self):
        """
        Create connections between step entities.
        """
        for node in self.root_nodes['SHAPE_REPRESENTATION_RELATIONSHIP']:
            # Associate each step representation entity to its SHAPE_REPRESENTATION
            function = self.functions[node]
            id_shape_representation = int(function.arg[2][1:])
            id_shape = int(function.arg[3][1:])
            self.connections[id_shape_representation].append(id_shape)
            self.functions[id_shape_representation].arg.append(f'#{id_shape}')
        for node in self.root_nodes['SHAPE_DEFINITION_REPRESENTATION']:
            # Associate each step representation entity to its SHAPE_REPRESENTATION
            function = self.functions[node]
            id_product_definition_shape = int(function.arg[0][1:])
            id_product_definition = int(self.functions[id_product_definition_shape].arg[2][1:])
            id_shape_representation = int(function.arg[1][1:])
            self.connections[id_product_definition].append(node)
            self.functions[id_product_definition].arg.append(f'#{node}')
            if self.functions[id_shape_representation].name == "SHAPE_REPRESENTATION" and \
                    len(self.functions[id_shape_representation].arg) >= 4:
                # todo: take all the "arg" starting from index 3 to end ??? needs investigation
                id_shapes = [int(arg[1:]) for arg in self.functions[id_shape_representation].arg[3:]]
                self.connections[id_product_definition].extend(id_shapes)
                for id_shape in id_shapes:
                    self.functions[id_product_definition].arg.append(f'#{id_shape}')
            elif self.functions[id_shape_representation].name in STEP_REPRESENTATION_ENTITIES:
                self.connections[id_product_definition].append(id_shape_representation)
                self.functions[id_product_definition].arg.append(f'#{id_shape_representation}')

            shell_node = self.shape_definition_representation_to_shell_node(node)
            product_node = self.shape_definition_representation_to_product_node(node)
            if shell_node:
                self.connections[shell_node].append(product_node)
                self.functions[shell_node].arg.append(f'#{product_node}')

        for node in self.root_nodes['CONTEXT_DEPENDENT_SHAPE_REPRESENTATION']:
            next_assembly_usage_occurrence = \
                self.context_dependent_shape_representation_to_next_assembly_usage_occurrence(node)
            self.connections[next_assembly_usage_occurrence].append(node)
            self.functions[next_assembly_usage_occurrence].arg.append(f'#{node}')

    def instatiate_assembly(self, object_dict):
        assemblies_structure, valid_entities = self.get_assembly_structure()

        instantiate_ids = list(assemblies_structure.keys())
        error = True
        last_error = None
        none_primitives = set()
        assembly_shape_ids = []
        while error:
            try:
                # here we invert instantiate_ids because if the code enter inside the except
                # block, we want to loop from the last KeyError to the first. This avoids an infinite loop
                for instantiate_id in reversed(instantiate_ids):
                    if instantiate_id in object_dict or instantiate_id in none_primitives:
                        instantiate_ids.pop()
                        continue
                    product_id = self.shape_definition_representation_to_product_node(instantiate_id)
                    name = self.functions[product_id].arg[0]
                    id_shape_representation = int(self.functions[instantiate_id].arg[1][1:])
                    ids_frames = self.functions[id_shape_representation].arg[1]
                    self.parse_arguments(ids_frames)
                    assembly_frame = object_dict[ids_frames[0]]

                    assembly_shape_ids, assembly_position_ids = self.get_assembly_data(
                        assemblies_structure[instantiate_id], valid_entities, assembly_frame, object_dict)
                    assembly_positions = []
                    list_primitives = []
                    for id_shape, id_frame in zip(assembly_shape_ids, assembly_position_ids):
                        if id_shape not in none_primitives:
                            assembly_positions.append(object_dict[id_frame])
                            list_primitives.append(object_dict[id_shape])

                    if not list_primitives:
                        none_primitives.add(instantiate_id)

                    volmdlr_object = volmdlr.core.Assembly(list_primitives, assembly_positions, assembly_frame,
                                                           name=name)
                    object_dict[instantiate_id] = volmdlr_object
                    last_error = None
                error = False
            except KeyError as key:
                # Sometimes the search don't instantiate the nodes of a
                # depth in the right order, leading to error
                if last_error == key.args[0]:
                    raise NotImplementedError('Error instantiating assembly') from key
                if key.args[0] in assembly_shape_ids:
                    instantiate_ids.append(key.args[0])

                last_error = key.args[0]
        return volmdlr_object

    def to_volume_model(self, show_times: bool = False):
        """
        Translate a step file into a volmdlr object.

        :param show_times: if True, displays how many times a given class has been
            instantiated and the total time of all the instantiations of this
            given class.
        :type show_times: bool
        :return: A volmdlr solid object.
        :rtype: :class:`volmdlr.core.VolumeModel`
        """
        object_dict = {}
        times = {}
        self.create_connections()
        root_nodes = self.root_nodes
        # ------------------------------------------------------
        # TODO: This isn't a 100% right. Each SHAPE_REPRESENTATION has its own geometric context
        geometric_representation_dict = root_nodes["GEOMETRIC_REPRESENTATION_CONTEXT"]
        geometric_representation_nodes = list(geometric_representation_dict.values())
        object_dict, times = self._helper_instantiate(geometric_representation_nodes[0],
                                                      object_dict, times, show_times)
        arguments = self.functions[geometric_representation_nodes[0]].arg[:]
        self.global_uncertainty = object_dict[int(arguments[1][0][1:])]
        self.length_conversion_factor = object_dict[int(arguments[2][0][1:])]
        self.angle_conversion_factor = object_dict[int(arguments[2][1][1:])]
        # ------------------------------------------------------
        shape_representations = root_nodes["SHAPE_REPRESENTATION"]
        nodes = self.create_node_list(shape_representations)
        errors = set()
        for node in nodes:
            if node is None:
                continue
            object_dict, times = self._helper_instantiate(node, object_dict, times, show_times)

            if not object_dict[node]:
                errors.add(node)

        if show_times:
            print()
            for key, value in times.items():
                print(f'| {key} : {value}')
            print()

        if self.root_nodes["NEXT_ASSEMBLY_USAGE_OCCURRENCE"]:
            return volmdlr.core.VolumeModel([self.instatiate_assembly(object_dict)])
        primitives = []
        shapes = [object_dict[shape] for shape in shape_representations]
        for shape in shapes:
            if isinstance(shape, list):
                primitives.extend(shape)
            else:
                primitives.append(shape)
        volume_model = volmdlr.core.VolumeModel(primitives)
        # volume_model = volmdlr.core.VolumeModel([object_dict[shell_node] for shell_node in shell_nodes])
        return volume_model

    def _helper_instantiate(self, node, object_dict, times, show_times):
        """
        Helper method to translate step entities into volmdlr objects.
        """
        instantiate_ids = [node]
        error = True
        while error:
            try:
                # here we invert instantiate_ids because if the code enter inside the except
                # block, we want to loop from the last KeyError to the first. This avoids an infinite loop
                for instantiate_id in instantiate_ids[::-1]:
                    t_tracker = time.time()
                    volmdlr_object = self.instantiate(
                        self.functions[instantiate_id].name,
                        self.functions[instantiate_id].arg[:], object_dict, instantiate_id)
                    t_tracker = time.time() - t_tracker
                    object_dict[instantiate_id] = volmdlr_object
                    if show_times:
                        if volmdlr_object.__class__ not in times:
                            times[volmdlr_object.__class__] = [1, t_tracker]
                        else:
                            times[volmdlr_object.__class__][0] += 1
                            times[volmdlr_object.__class__][1] += t_tracker
                error = False
            except KeyError as key:
                # Sometimes the search don't instantiate the nodes of a
                # depth in the right order, leading to error
                instantiate_ids.append(key.args[0])

        return object_dict, times

    def to_points(self):
        """Returns a list containing all the points present in a step file."""
        object_dict = {}
        points3d = []
        for stepfunction in self.functions.values():
            if stepfunction.name == 'CARTESIAN_POINT':
                # INSTANTIATION
                name = self.functions[stepfunction.id].name
                arguments = self.functions[stepfunction.id].arg[:]
                self.parse_arguments(arguments)
                if arguments[1].count(',') == 2:
                    volmdlr_object = STEP_TO_VOLMDLR[name].from_step(
                        arguments, object_dict)
                    points3d.append(volmdlr_object)

        # remove first point because it refers to origin
        return points3d[1:]

    def plot_data(self):
        graph = self.graph().copy()

        graph.remove_nodes_from([stepfunction.id for stepfunction
                                 in self.functions.values()
                                 if stepfunction.name in ['CARTESIAN_POINT', 'DIRECTION']])
        return [plot_data.graph.NetworkxGraph(graph=graph)]


@dataclass
class StepReaderReport:
    """
    Data class to save a report after translating a step file to volmdlr object.
    """
    step_name: str = " "
    total_number_of_faces: int = 0
    faces_read: int = 0
    sucess_rate: float = 0.0
    errors: list = field(default_factory=list)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Common abstract primitives.

"""

import math
from typing import Dict, List

from numpy import zeros
from scipy.optimize import linprog

# import dessia_common as dc
import volmdlr.edges


class RoundedLineSegments:
    """
    Rounded Line Segments class.

    """
    _non_serializable_attributes = ['line_class', 'arc_class', 'basis_primitives', 'primitives']

    line_class = volmdlr.edges.LineSegment
    arc_class = volmdlr.edges.ArcMixin

    def __init__(self, points: List[volmdlr.Point3D], radius: Dict[str, float],
                 closed: bool = False, adapt_radius: bool = False, name: str = ''):

        self.points = points
        self.radius = {int(k): v for k, v in radius.items()}
        self.closed = closed
        self.adapt_radius = adapt_radius
        self.name = name
        self.npoints = len(points)

    def get_points(self, point_index):
        """
        Gets points to calculate the arc features.
        """
        if self.closed:
            if point_index == 0:
                pt1 = self.points[-1]
            else:
                pt1 = self.points[point_index - 1]
            pti = self.points[point_index]
            if point_index < self.npoints - 1:
                pt2 = self.points[point_index + 1]
            else:
                pt2 = self.points[0]
        else:
            pt1 = self.points[point_index - 1]
            pti = self.points[point_index]
            pt2 = self.points[point_index + 1]
        return pt1, pti, pt2

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new RoundedLineSegments.

        side = 'old' or 'new'
        """
        return self.__class__([point.frame_mapping(frame, side)
                               for point in self.points], radius=self.radius,
                              adapt_radius=self.adapt_radius,
                              name=self.name)

    def arc_features(self, point_index: int):
        raise NotImplementedError('The method arc_features should be overloaded.')

    def _primitives(self):
        alpha = {}
        dist = {}
        lines_length = {}
        # Computing optimal radii
        rounded_points_indices = [int(i) for i in sorted(self.radius.keys())]
        groups = []
        arcs = {}

        if self.radius != {}:
            group = [rounded_points_indices[0]]
            _, _, _, dist0, alpha0 = self.arc_features(rounded_points_indices[0])
            dist[rounded_points_indices[0]] = dist0
            alpha[rounded_points_indices[0]] = alpha0

            for i in rounded_points_indices[1:]:
                # Computing the arc
                _, _, _, dist2, alpha2 = self.arc_features(i)
                dist[i] = dist2
                alpha[i] = alpha2
                if i - 1 in self.radius:
                    point1 = self.points[i - 1]
                    point2 = self.points[i]
                    length = (point2 - point1).norm()
                    lines_length[i - 1] = length
                    dist1 = dist[i - 1]

                    if dist1 + dist2 <= length:
                        groups.append(group)
                        group = [i]
                    else:
                        if not self.adapt_radius:
                            raise ValueError
                        group.append(i)
                else:
                    if group:
                        groups.append(group)
                    group = [i]
            if group:
                groups.append(group)
            if self.adapt_radius:
                if self.closed:
                    if 0 in groups[0]:
                        if self.npoints in groups[-1]:
                            new_group = groups[0] + groups[-1]
                            groups[0] = new_group
                            del groups[-1]

                groups2 = []
                ndof = 0
                dof = {}
                neq_ub = 0
                bounds = []
                for group in groups:
                    len_group = len(group)
                    if len_group == 1:
                        # Single point, reducing its radius by simple computation if needed
                        ipoint = group[0]
                        if self.closed:
                            if ipoint == 0:
                                point1 = self.points[-1]
                                point2 = self.points[0]
                                point3 = self.points[1]
                            elif ipoint == self.npoints - 1:
                                point1 = self.points[-2]
                                point2 = self.points[-1]
                                point3 = self.points[0]
                            else:
                                point1 = self.points[ipoint - 1]
                                point2 = self.points[ipoint]
                                point3 = self.points[ipoint + 1]

                        else:
                            point1 = self.points[ipoint - 1]
                            point2 = self.points[ipoint]
                            point3 = self.points[ipoint + 1]

                        distance_1 = point1.point_distance(point2)
                        distance_2 = point2.point_distance(point3)

                        if dist[ipoint] > (min(distance_1, distance_2)):
                            self.radius[ipoint] = min(self.radius[ipoint],
                                                      min(distance_1, distance_2) * math.tan(alpha[ipoint]))

                    else:
                        # Adding to dof
                        bounds.extend([(0, self.radius[j] / math.tan(alpha[j])) for j in group])
                        dof.update({j: ndof + i for i, j in enumerate(group)})
                        ndof += len_group
                        groups2.append(group)
                        neq_ub += len_group - 1

                # Constructing simplex problem
                # C matrix:
                if ndof > 0:
                    c = zeros(ndof)
                    for j, i in dof.items():
                        c[i] = -math.tan(alpha[j])

                    a_ub = zeros((neq_ub, ndof))
                    b_ub = zeros(neq_ub)
                    ieq_ub = 0

                    for group in groups2:
                        for ip1, ip2 in zip(group[:-1], group[1:]):
                            a_ub[ieq_ub, dof[ip1]] = 1
                            a_ub[ieq_ub, dof[ip2]] = 1
                            b_ub[ieq_ub] = lines_length[ip1]
                            ieq_ub += 1
                    optimized_radius_solution = linprog(c, a_ub, b_ub, bounds=bounds)

                    for ipoint, dof_point in dof.items():
                        radius = optimized_radius_solution .x[dof_point] * math.tan(alpha[ipoint])
                        if radius > 1e-10:
                            self.radius[ipoint] = radius
                        else:
                            del self.radius[ipoint]

            # Creating geometry
            # Creating arcs
            for ipoint, radius in self.radius.items():
                p_start, p_iterior, p_end, _, _ = self.arc_features(ipoint)
                arcs[ipoint] = self.arc_class.from_3_points(p_start, p_iterior, p_end)

        return self.primitives_from_arcs(arcs)

    def primitives_from_arcs(self, arcs):
        primitives = []
        # Creating lines
        for index_line in range(self.npoints - 1):
            if index_line in self.radius:
                arc1 = arcs[index_line]
                primitives.append(arc1)
                if index_line + 1 in self.radius:
                    arc2 = arcs[index_line + 1]
                    if not arc1.end.is_close(arc2.start):
                        primitives.append(self.line_class(arc1.end, arc2.start))
                else:
                    if not arc1.end.is_close(self.points[index_line + 1]):
                        primitives.append(self.line_class(arc1.end, self.points[index_line + 1]))
            else:
                p1 = self.points[index_line]
                if index_line + 1 in self.radius:
                    arc2 = arcs[index_line + 1]
                    if not p1.is_close(arc2.start):
                        primitives.append(self.line_class(p1, arc2.start))
                else:
                    primitives.append(self.line_class(p1, self.points[index_line + 1]))

        if self.closed:
            if self.npoints - 1 in self.radius:
                arc1 = arcs[self.npoints - 1]
                primitives.append(arc1)
                if 0 in self.radius:
                    arc2 = arcs[0]
                    if not arc1.end.is_close(arc2.start):
                        primitives.append(self.line_class(arc1.end, arc2.start))
                else:
                    primitives.append(self.line_class(arc1.end, self.points[0]))
            else:
                p1 = self.points[self.npoints - 1]
                if 0 in self.radius:
                    arc2 = arcs[0]
                    if not p1.is_close(arc2.start):
                        primitives.append(self.line_class(p1, arc2.start))
                else:
                    primitives.append(self.line_class(p1, self.points[0]))

        return primitives



"""volmdlr shells module."""
import math
import random
import traceback
import warnings
from itertools import chain, product
from typing import Any, Dict, Iterable, List, Tuple, Union

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from dessia_common.core import DessiaObject
from dessia_common.typings import JsonSerializable
from numpy.typing import NDArray
from trimesh import Trimesh

import volmdlr.core
import volmdlr.core_compiled
import volmdlr.faces
import volmdlr.geometry
from volmdlr import curves, display, edges, surfaces, wires
from volmdlr.core import (edge_in_list, get_edge_index_in_list,
                          get_point_index_in_list, point_in_list)
from volmdlr.utils.step_writer import (geometric_context_writer,
                                       product_writer, step_ids_to_str)


def union_list_of_shells(list_shells):
    """
    Perform union operations with all shells from a list, until all groups of adjacent shells are merged.

    :param list_shells: A list of shells
    :return: A list of merged shells
    """
    shells = []
    union_shells = [list_shells.pop(0)]
    len_previous_shells_list = len(list_shells)
    while True:
        if not list_shells and union_shells:
            list_shells = union_shells[:]
            len_previous_shells_list = len(list_shells)
            union_shells = [list_shells.pop(0)]

        for i, union_shell in enumerate(union_shells[:]):
            for j, octant_box in enumerate(list_shells[:]):
                union = union_shell.union(octant_box)
                if len(union) != 2:
                    union_shells.pop(i)
                    union[0].merge_faces()
                    union_shells.insert(0, union[0])
                    list_shells.pop(j)
                    break
                if len(list_shells) != len_previous_shells_list - 1:
                    union_shells.insert(0, union[1])
                    list_shells.remove(octant_box)
                    break
            else:
                shells.append(union_shell)
                union_shells.pop(i)
                if not union_shells and list_shells:
                    union_shells = [list_shells.pop(0)]
                break
            break
        if not union_shells:
            break
    return shells


class Shell3D(volmdlr.core.CompositePrimitive3D):
    """
    A 3D shell composed of multiple faces.

    This class represents a 3D shell, which is a collection of connected
    faces with volume. It is a subclass of the `CompositePrimitive3D` class
    and inherits all of its attributes and methods.


    :param faces: The faces of the shell.
    :type faces: List[`Face3D`]
    :param color: The color of the shell.
    :type color: Tuple[float, float, float]
    :param alpha: The transparency of the shell, should be a value in the interval (0, 1).
    :type alpha: float
    :param name: The name of the shell.
    :type name: str
    :param bounding_box: The bounding box of the shell.
    :type bounding_box: :class:`volmdlr.core.BoundingBox`
    """
    _standalone_in_db = True
    _non_serializable_attributes = ['primitives']
    _non_data_eq_attributes = ['name', 'color', 'alpha', 'bounding_box', 'primitives']
    _non_data_hash_attributes = []
    STEP_FUNCTION = None

    def __init__(self, faces: List[volmdlr.faces.Face3D],
                 color: Tuple[float, float, float] = None,
                 alpha: float = 1.,
                 name: str = '',
                 bounding_box: volmdlr.core.BoundingBox = None):

        self.faces = faces
        if not color:
            self.color = volmdlr.core.DEFAULT_COLOR
        else:
            self.color = color
        self.alpha = alpha
        self._bbox = None
        if bounding_box:
            self._bbox = bounding_box

        self._faces_graph = None
        self._vertices_graph = None
        self._vertices_points = None
        self._shell_octree_decomposition = None

        volmdlr.core.CompositePrimitive3D.__init__(self,
                                                   primitives=faces, color=color, alpha=alpha,
                                                   name=name)

    def _data_hash(self):
        return len(self.faces)  # sum(face._data_hash() for face in self.faces)

    def _data_eq(self, other_object):
        if other_object.__class__.__name__ != self.__class__.__name__:
            return False
        for face1, face2 in zip(self.faces, other_object.faces):
            if not face1._data_eq(face2):
                return False
        return True

    def __eq__(self, other_face):
        if not self.__class__ == other_face.__class__:
            return False
        if len(self.faces) != len(other_face.faces):
            return False
        if all(face in self.faces for face in other_face.faces):
            return True
        return False

    def __hash__(self):
        return hash((self.__class__.__name__, tuple(self.faces)))

    @staticmethod
    def _helper_getter_vertices_points(faces):
        """Helper method to get shells faces vertices."""
        vertices_points = []
        for face in faces:
            for contour in [face.outer_contour3d] + face.inner_contours3d:
                for edge in contour.primitives:
                    if not volmdlr.core.point_in_list(edge.start, vertices_points):
                        vertices_points.append(edge.start)
                    if not volmdlr.core.point_in_list(edge.end, vertices_points):
                        vertices_points.append(edge.end)
        return vertices_points

    @property
    def vertices_points(self):
        """Gets the shell's vertices points. """
        if self._vertices_points is None:
            self._vertices_points = self._helper_getter_vertices_points(self.faces)
        return self._vertices_points

    @property
    def vertices_graph(self):
        """
        Gets the shells faces graph using networkx.

        :return: return a networkx graph for a shell faces' vertices.
        """
        if not self._faces_graph:
            vertices_graph = nx.Graph()
            for face in self.faces:
                for edge in face.outer_contour3d.primitives:
                    edge_start_index = volmdlr.core.get_point_index_in_list(edge.start, self.vertices_points)
                    edge_end_index = volmdlr.core.get_point_index_in_list(edge.end, self.vertices_points)
                    vertices_graph.add_edge(edge_start_index, edge_end_index, edge=edge)
            self._vertices_graph = vertices_graph
        return self._vertices_graph

    @staticmethod
    def _faces_graph_search_bridges(faces, graph, components, face_vertices, vertices_points):
        """
        Search for neighboring faces in the connected components to fix the graph.

        To make the shell faces' topology graph, we search faces that share the same vertices.
        Sometimes, in very specific cases, it can occur that two faces are neighbors of each other
        but their vertices aren't coincident.

        This method tries to find those cases, by searching bridges, which are edges that connect two components.
        The search is performed by checking if any vertices of one face in one component lie on the contours of
        another face in the other component. If such a connection is found, the two faces are considered neighbors.
        Once a connection is found between a pair of components we break the execution and proceed to
        the next pair of components. This behavior, in the average, avoids the need to verify all possible combinations

        :param graph: The graph representing the faces' topology.
        :type graph: nx.Graph
        :param components: A list of sets, where each set contains the indices of faces in a connected component.
        :type components: list
        :param face_vertices: A dictionary mapping face indices to their corresponding vertex indices.
        :type face_vertices: dict
        :param vertices_points: A list containing all faces vertices points.
        :type vertices_points: list[Point3D].
        :return: The updated graph with no disconnected components.
        """
        stack = components.copy()
        found = False

        def check_faces(face, other_face_id):
            for point_id in face_vertices[other_face_id]:
                point = vertices_points[point_id]
                if face.outer_contour3d.point_over_contour(point):
                    return True
                if face.inner_contours3d:
                    for inner_contour in face.inner_contours3d:
                        if inner_contour.point_over_contour(point):
                            return True
            return False

        while stack:
            group = stack.pop(0)
            for face_id_i in group:
                face_i = faces[face_id_i]
                for other_group in stack:
                    for face_id_j in other_group:
                        if check_faces(face_i, face_id_j):
                            found = True
                            graph.add_edge(face_id_i, face_id_j)
                            for face_id_k in graph.neighbors(face_id_j):
                                if check_faces(face_i, face_id_k):
                                    graph.add_edge(face_id_i, face_id_k)
                        if found:
                            break
        return graph

    @staticmethod
    def _helper_create_faces_graph(faces, vertices_points=None, verify_connected_components=True):
        if vertices_points is None:
            vertices_points = Shell3D._helper_getter_vertices_points(faces)

        graph = nx.Graph()
        vertice_faces = {}
        face_vertices = {}

        for face_index, face in enumerate(faces):
            face_contour_primitives = face.outer_contour3d.primitives
            for inner_contour in face.inner_contours3d:
                face_contour_primitives.extend(inner_contour.primitives)
            for edge in face_contour_primitives:
                start_index = volmdlr.core.get_point_index_in_list(edge.start, vertices_points)
                vertice_faces.setdefault(start_index, set()).add(face_index)
                face_vertices.setdefault(face_index, set()).add(start_index)

        for i, _ in enumerate(faces):
            face_i_vertices = face_vertices[i]
            for vertice in face_i_vertices:
                connected_faces = vertice_faces[vertice]
                connected_faces.discard(i)
                for j in connected_faces:
                    graph.add_edge(i, j)

        if verify_connected_components:
            components = list(nx.connected_components(graph))
            if len(components) > 1:
                graph = Shell3D._faces_graph_search_bridges(faces, graph, components, face_vertices, vertices_points)
        return graph

    def faces_graph(self, verify_connected_components=True):
        """
        Gets the shells faces topology graph using networkx.

        :return: return a networkx graph for a shell faces.
        """
        if not self._faces_graph:
            self._faces_graph = self._helper_create_faces_graph(
                self.faces, self.vertices_points, verify_connected_components)
        return self._faces_graph

    def to_dict(self, *args, **kwargs):
        """
        Serializes a 3-dimensional open shell into a dictionary.

        This method does not use pointers for faces as it has no sense
        to have duplicate faces.

        :return: A serialized version of the OpenShell3D
        :rtype: dict

        .. see also::
            How `serialization and de-serialization`_ works in dessia_common

        .. _serialization and deserialization:
        https://documentation.dessia.tech/dessia_common/customizing.html#overloading-the-dict-to-object-method

        """
        dict_ = DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'faces': [f.to_dict(use_pointers=False) for f in self.faces]})
        if self._bbox:
            dict_['bounding_box'] = self._bbox.to_dict()

        return dict_

    @classmethod
    def from_step(cls, arguments, object_dict, **kwargs):
        """
        Converts a step primitive to a Open Shell 3D.

        :param arguments: The arguments of the step primitive.
        :type arguments: list
        :param object_dict: The dictionary containing all the step primitives
            that have already been instantiated.
        :type object_dict: dict
        :return: The corresponding OpenShell3D object.
        :rtype: :class:`volmdlr.faces.OpenShell3D`
        """
        # Quick fix:
        # ----------------------------------
        name = arguments[0][1:-1]
        if isinstance(arguments[-1], int):
            step_product = object_dict[arguments[-1]]
            name = step_product[1:-1]
        # ----------------------------------
        faces = [object_dict[int(face[1:])] for face in arguments[1] if object_dict[int(face[1:])]]
        if not faces:
            return None
        return cls(faces, name=name)

    def to_step(self, current_id):
        """
        Creates step file entities from volmdlr objects.
        """
        step_content = ''
        faces_content = ''
        face_ids = []

        for face in self.faces:
            if isinstance(face, (volmdlr.faces.Face3D, surfaces.Surface3D)):
                face_content, face_sub_ids = face.to_step(current_id)
            else:
                face_content, face_sub_ids = face.to_step(current_id)
                face_sub_ids = [face_sub_ids]
            faces_content += face_content
            face_ids.extend(face_sub_ids)
            current_id = max(face_sub_ids)
        step_content += faces_content

        shell_id = current_id + 1

        step_content += f"#{shell_id} = {self.STEP_FUNCTION}('{self.name}'," \
                        f"({step_ids_to_str(face_ids)}));\n"
        manifold_id = shell_id + 1
        if self.STEP_FUNCTION == "CLOSED_SHELL":
            step_content += f"#{manifold_id} = MANIFOLD_SOLID_BREP('{self.name}',#{shell_id});\n"
        else:
            step_content += f"#{manifold_id} = SHELL_BASED_SURFACE_MODEL('{self.name}',(#{shell_id}));\n"

        return step_content, manifold_id

    def to_step_product(self, current_id):
        """
        Creates step file entities from volmdlr objects.
        """
        step_content = ''
        faces_content = ''
        face_ids = []

        product_content, shape_definition_repr_id = product_writer(current_id, self.name)
        shape_representation_id = shape_definition_repr_id + 1
        product_id = shape_definition_repr_id - 4
        product_definition_id = shape_definition_repr_id - 2
        step_content += product_content

        brep_id = shape_representation_id
        frame_content, frame_id = volmdlr.OXYZ.to_step(brep_id)
        manifold_id = frame_id + 1
        shell_id = manifold_id + 1
        current_id = shell_id + 1
        for face in self.faces:
            if isinstance(face, (volmdlr.faces.Face3D, surfaces.Surface3D)):
                face_content, face_sub_ids = face.to_step(current_id)
            else:
                face_content, face_sub_ids = face.to_step(current_id)
                face_sub_ids = [face_sub_ids]
            faces_content += face_content
            face_ids.extend(face_sub_ids)
            current_id = max(face_sub_ids)

        geometric_context_content, geometric_representation_context_id = geometric_context_writer(current_id)

        if self.STEP_FUNCTION == "CLOSED_SHELL":
            step_content += f"#{brep_id} = ADVANCED_BREP_SHAPE_REPRESENTATION('',(#{frame_id},#{manifold_id})," \
                            f"#{geometric_representation_context_id});\n"
            step_content += frame_content
            step_content += f"#{manifold_id} = MANIFOLD_SOLID_BREP('{self.name}',#{shell_id});\n"
        else:
            step_content += f"#{brep_id} = MANIFOLD_SURFACE_SHAPE_REPRESENTATION('',(#{frame_id},#{manifold_id})," \
                            f"#{geometric_representation_context_id});\n"
            step_content += frame_content
            step_content += f"#{manifold_id} = SHELL_BASED_SURFACE_MODEL('{self.name}',(#{shell_id}));\n"

        step_content += f"#{shell_id} = {self.STEP_FUNCTION}('{self.name}'," \
                        f"({step_ids_to_str(face_ids)}));\n"
        step_content += faces_content

        step_content += geometric_context_content

        product_related_category = geometric_representation_context_id + 1
        step_content += f"#{product_related_category} = PRODUCT_RELATED_PRODUCT_CATEGORY(" \
                        f"'part',$,(#{product_id}));\n"
        draughting_id = product_related_category + 1
        step_content += f"#{draughting_id} = DRAUGHTING_PRE_DEFINED_CURVE_FONT('continuous');\n"
        color_id = draughting_id + 1
        primitive_color = (1, 1, 1)
        if hasattr(self, 'color') and self.color is not None:
            primitive_color = self.color
        step_content += f"#{color_id} = COLOUR_RGB('',{round(float(primitive_color[0]), 4)}," \
                        f"{round(float(primitive_color[1]), 4)},{round(float(primitive_color[2]), 4)});\n"

        curve_style_id = color_id + 1
        step_content += f"#{curve_style_id} = CURVE_STYLE('',#{draughting_id}," \
                        f"POSITIVE_LENGTH_MEASURE(0.1),#{color_id});\n"

        fill_area_color_id = curve_style_id + 1
        step_content += f"#{fill_area_color_id} = FILL_AREA_STYLE_COLOUR('',#{color_id});\n"

        fill_area_id = fill_area_color_id + 1
        step_content += f"#{fill_area_id} = FILL_AREA_STYLE('',#{fill_area_color_id});\n"

        suface_fill_area_id = fill_area_id + 1
        step_content += f"#{suface_fill_area_id} = SURFACE_STYLE_FILL_AREA(#{fill_area_id});\n"

        suface_side_style_id = suface_fill_area_id + 1
        step_content += f"#{suface_side_style_id} = SURFACE_SIDE_STYLE('',(#{suface_fill_area_id}));\n"

        suface_style_usage_id = suface_side_style_id + 1
        step_content += f"#{suface_style_usage_id} = SURFACE_STYLE_USAGE(.BOTH.,#{suface_side_style_id});\n"

        presentation_style_id = suface_style_usage_id + 1

        step_content += f"#{presentation_style_id} = PRESENTATION_STYLE_ASSIGNMENT((#{suface_style_usage_id}," \
                        f"#{curve_style_id}));\n"

        styled_item_id = presentation_style_id + 1
        if self.__class__.__name__ == 'OpenShell3D':
            for face_id in face_ids:
                step_content += f"#{styled_item_id} = STYLED_ITEM('color',(#{presentation_style_id})," \
                                f"#{face_id});\n"
                styled_item_id += 1
            styled_item_id -= 1
        else:
            step_content += f"#{styled_item_id} = STYLED_ITEM('color',(#{presentation_style_id})," \
                            f"#{manifold_id});\n"
        mechanical_design_id = styled_item_id + 1
        step_content += f"#{mechanical_design_id} =" \
                        f" MECHANICAL_DESIGN_GEOMETRIC_PRESENTATION_REPRESENTATION(" \
                        f"'',(#{styled_item_id}),#{geometric_representation_context_id});\n"
        current_id = mechanical_design_id

        return step_content, current_id, [brep_id, product_definition_id]

    def to_step_face_ids(self, current_id):
        """
        Creates step file entities from volmdlr objects.
        """
        step_content = ''
        face_ids = []
        for face in self.faces:
            if isinstance(face, volmdlr.faces.Face3D):
                face_content, face_sub_ids = face.to_step(current_id)
            else:
                face_content, face_sub_ids = face.to_step(current_id)
                face_sub_ids = [face_sub_ids]
            step_content += face_content
            face_ids.extend(face_sub_ids)
            current_id = max(face_sub_ids) + 1

        shell_id = current_id
        step_content += f"#{current_id} = {self.STEP_FUNCTION}('{self.name}'," \
                        f"({volmdlr.core.step_ids_to_str(face_ids)}));\n"
        manifold_id = shell_id + 1
        step_content += f"#{manifold_id} = SHELL_BASED_SURFACE_MODEL('{self.name}',(#{shell_id}));\n"

        frame_content, frame_id = volmdlr.OXYZ.to_step(manifold_id + 1)
        step_content += frame_content
        brep_id = frame_id + 1
        step_content += f"#{brep_id} = MANIFOLD_SURFACE_SHAPE_REPRESENTATION('',(#{frame_id},#{manifold_id}),#7);\n"

        return step_content, brep_id, face_ids

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        Open Shell 3D / Closed Shell 3D rotation.

        :param center: rotation center.
        :param axis: rotation axis.
        :param angle: angle rotation.
        :return: a new rotated OpenShell3D.
        """
        new_faces = [face.rotation(center, axis, angle) for face
                     in self.faces]
        return self.__class__(new_faces, color=self.color, alpha=self.alpha, name=self.name)

    def translation(self, offset: volmdlr.Vector3D):
        """
        Shell3D translation.

        :param offset: translation vector.
        :return: A new translated Open Shell 3D.
        """
        new_faces = [face.translation(offset) for face in
                     self.faces]
        return self.__class__(new_faces, color=self.color, alpha=self.alpha,
                              name=self.name)

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new OpenShell3D.

        side = 'old' or 'new'.
        """
        new_faces = [face.frame_mapping(frame, side) for face in
                     self.faces]
        return self.__class__(new_faces, name=self.name)

    def copy(self, deep=True, memo=None):
        """
        Copy of Shell3D.

        :return: return a copy a shell 3D.
        """
        new_faces = [face.copy(deep=deep, memo=memo) for face in self.faces]
        return self.__class__(new_faces, color=self.color, alpha=self.alpha,
                              name=self.name)

    @property
    def bounding_box(self):
        """
        Returns the boundary box.

        """
        if not self._bbox:
            self._bbox = self.get_bounding_box()
        return self._bbox

    @bounding_box.setter
    def bounding_box(self, new_bounding_box):
        self._bbox = new_bounding_box

    def get_bounding_box(self):
        """Gets the Shell bounding box."""
        bounding_boxes = []
        for face in self.faces:
            if face.outer_contour3d.primitives:
                bounding_boxes.append(face.bounding_box)
        return volmdlr.core.BoundingBox.from_bounding_boxes(bounding_boxes)

    def cut_by_plane(self, plane_3d: surfaces.Plane3D):
        """
        Cut Shell3D by plane 3d.

        :param plane_3d: plane 3d o cut shell.
        :return: return a list of faces containing the shell's sections at the plane 3d given.
        """
        frame_block = self.bounding_box.to_frame()
        frame_block.u = 1.1 * frame_block.u
        frame_block.v = 1.1 * frame_block.v
        frame_block.w = 1.1 * frame_block.w
        block = volmdlr.primitives3d.Block(frame_block,
                                           color=(0.1, 0.2, 0.2),
                                           alpha=0.6)
        face_3d = block.cut_by_orthogonal_plane(plane_3d)
        intersection_primitives = []
        for face in self.faces:
            intersection_wires = face.face_intersections(face_3d)
            if intersection_wires:
                for intersection_wire in intersection_wires:
                    intersection_primitives.extend(intersection_wire.primitives)
        contours3d = wires.Contour3D.contours_from_edges(
            intersection_primitives[:])
        if not contours3d:
            return []
        contours2d = [contour.to_2d(plane_3d.frame.origin,
                                    plane_3d.frame.u,
                                    plane_3d.frame.v) for contour in contours3d]
        resulting_faces = []
        for contour2d in contours2d:
            if contour2d.area() > 1e-7:
                surface2d = surfaces.Surface2D(contour2d, [])
                resulting_faces.append(volmdlr.faces.PlaneFace3D(plane_3d, surface2d))
        return resulting_faces

    def linesegment_intersections(self, linesegment3d: edges.LineSegment3D) \
            -> List[Tuple[volmdlr.faces.Face3D, List[volmdlr.Point3D]]]:
        """
        Gets the intersections of a Shell3D with a Line Segment 3D.

        :param linesegment3d: other line segment.
        :return: List of tuples (face, intersections).
        """
        # intersections = []
        for face in self.faces:
            face_intersections = face.linesegment_intersections(linesegment3d)
            if face_intersections:
                yield face, face_intersections
        # return intersections

    def line_intersections(self,
                           line3d: curves.Line3D) \
            -> List[Tuple[volmdlr.faces.Face3D, List[volmdlr.Point3D]]]:
        """
        Gets the intersections of a Shell3D with a Line Segment 3D.

        :param line3d: other line segment.
        :return: List of tuples (face, intersections).
        """
        intersections = []
        for face in self.faces:
            face_intersections = face.line_intersections(line3d)
            if face_intersections:
                intersections.append((face, face_intersections))
        return intersections

    def minimum_distance_points(self, shell2):
        """
        Returns a Measure object if the distance is not zero, otherwise returns None.

        """
        shell2_inter = self.is_intersecting_with(shell2)
        if shell2_inter:
            return None
        _, point1, point2 = self.minimum_distance(shell2, return_points=True)
        return point1, point2

    def minimum_distance_point(self,
                               point: volmdlr.Point3D) -> volmdlr.Point3D:
        """
        Computes the distance of a point to a Shell3D, whether it is inside or outside the Shell3D.

        """
        distance_min, point1_min = self.faces[0].distance_to_point(point,
                                                                   return_other_point=True)
        for face in self.faces[1:]:
            bbox_distance = self.bounding_box.distance_to_point(point)
            if bbox_distance < distance_min:
                distance, point1 = face.distance_to_point(point,
                                                          return_other_point=True)
                if distance < distance_min:
                    distance_min, point1_min = distance, point1

        return point1_min

    def minimum_distance(self, other_shell, return_points=False):
        shell_decomposition1 = self.shell_decomposition()
        shell_decomposition2 = other_shell.shell_decomposition()
        list_set_points1 = [{point for face in faces1
                             for point in face.outer_contour3d.discretization_points(number_points=10)} for _, faces1 in
                            shell_decomposition1.items()]
        list_set_points1 = [np.array([(point[0], point[1], point[2]) for point in sets_points1]) for sets_points1 in
                            list_set_points1]
        list_set_points2 = [{point for face in faces2
                             for point in face.outer_contour3d.discretization_points(number_points=10)} for _, faces2 in
                            shell_decomposition2.items()]
        list_set_points2 = [np.array([(point[0], point[1], point[2]) for point in sets_points2]) for sets_points2 in
                            list_set_points2]

        minimum_distance = math.inf
        index1, index2 = None, None
        for sets_points1, sets_points2 in product(list_set_points1, list_set_points2):
            distances = np.linalg.norm(sets_points2[:, np.newaxis] - sets_points1, axis=2)
            sets_min_dist = np.min(distances)
            if sets_min_dist < minimum_distance:
                minimum_distance = sets_min_dist
                index1 = next((i for i, x in enumerate(list_set_points1) if np.array_equal(x, sets_points1)), -1)
                index2 = next((i for i, x in enumerate(list_set_points2) if np.array_equal(x, sets_points2)), -1)
        faces1 = list(shell_decomposition1.values())[index1]
        faces2 = list(shell_decomposition2.values())[index2]

        minimum_distance = math.inf
        best_distance_points = None
        for face1, face2 in product(faces1, faces2):
            distance, point1, point2 = face1.face_minimum_distance(face2, True)
            if distance < minimum_distance:
                minimum_distance = distance
                best_distance_points = [point1, point2]
        if return_points:
            return minimum_distance, *best_distance_points
        return minimum_distance

    def shell_decomposition(self):
        if not self._shell_octree_decomposition:
            self._shell_octree_decomposition = volmdlr.faces.octree_decomposition(self.bounding_box, self.faces)
        return self._shell_octree_decomposition

    def intersection_internal_aabb_volume(self, shell2: 'OpenShell3D',
                                          resolution: float):
        """
        Aabb made of the intersection points and the points of self internal to shell2.
        """
        intersections_points = []
        for face1 in self.faces:
            for face2 in shell2.faces:
                intersection_points = face1.face_intersections(face2)
                if intersection_points:
                    intersection_points = [
                        intersection_points[0].primitives[0].start,
                        intersection_points[0].primitives[0].end]
                    intersections_points.extend(intersection_points)

        shell1_points_inside_shell2 = []
        for face in self.faces:
            for point in face.outer_contour3d.discretization_points(angle_resolution=resolution):
                if shell2.point_belongs(point):
                    shell1_points_inside_shell2.append(point)

        if len(intersections_points + shell1_points_inside_shell2) == 0:
            return 0
        bbox = volmdlr.core.BoundingBox.from_points(
            intersections_points + shell1_points_inside_shell2)
        return bbox.volume()

    def intersection_external_aabb_volume(self, shell2: 'OpenShell3D',
                                          resolution: float):
        """
        Aabb made of the intersection points and the points of self external to shell2.
        """
        intersections_points = []
        for face1 in self.faces:
            for face2 in shell2.faces:
                intersection_points = face1.face_intersections(face2)
                if intersection_points:
                    intersection_points = [
                        intersection_points[0].primitives[0].start,
                        intersection_points[0].primitives[0].end]
                    intersections_points.extend(intersection_points)

        shell1_points_outside_shell2 = []
        for face in self.faces:
            for point in face.outer_contour3d.discretization_points(
                    angle_resolution=resolution):
                if not shell2.point_belongs(point):
                    shell1_points_outside_shell2.append(point)

        if len(intersections_points + shell1_points_outside_shell2) == 0:
            return 0
        bbox = volmdlr.core.BoundingBox.from_points(
            intersections_points + shell1_points_outside_shell2)
        return bbox.volume()

    def face_on_shell(self, face):
        """
        Verifies if a face lies on the shell's surface.

        """
        for face_ in self.faces:
            if face_.face_inside(face):
                return True
        return False

    def point_on_shell(self, point: volmdlr.Point3D):
        """
        Verify if a point is on the shell (on one of the shell's faces).

        :param point: point to be verified.
        :return: return True or False.
        """
        for face in self.faces:
            if face.point_belongs(point):
                return True
        return False

    def triangulation(self):
        """
        Triangulation of a Shell3D.

        """
        meshes = []
        for i, face in enumerate(self.faces):
            try:
                face_mesh = face.triangulation()

            except Exception:
                face_mesh = None
                warnings.warn(f"Could not triangulate {face.__class__.__name__} with index {i} in the shell "
                              f"{self.name} faces. Probably because topology error in contour2d.")
                print(traceback.format_exc())
                continue
            if face_mesh:
                meshes.append(face_mesh)
        return display.DisplayMesh3D.merge_meshes(meshes)

    def babylon_meshes(self, merge_meshes=True):
        """
        Returns the babylonjs mesh.
        """
        if merge_meshes:
            return super().babylon_meshes()
        babylon_meshes = []
        for face in self.faces:
            face_babylon_meshes = face.babylon_meshes()
            if not face_babylon_meshes:
                continue
            if face_babylon_meshes[0]['positions']:
                babylon_meshes.extend(face.babylon_meshes())
        babylon_mesh = {'primitives_meshes': babylon_meshes}
        babylon_mesh.update(self.babylon_param())
        return [babylon_mesh]

    def plot(self, ax=None, color: str = 'k', alpha: float = 1.0):
        """
        Plot a Shell 3D using Matplotlib.

        """
        if ax is None:
            ax = plt.figure().add_subplot(111, projection='3d')

        for face in self.faces:
            face.plot(ax=ax, color=color, alpha=alpha)

        return ax

    def project_coincident_faces_of(self, shell):
        """
        Divides self's faces based on coincident shell's faces.

        """

        list_faces = []
        initial_faces = self.faces[:]

        for face1 in initial_faces:
            list_faces.extend(face1.project_faces(shell.faces))

        return self.__class__(list_faces)

    def get_geo_lines(self, update_data,
                      point_mesh_size: float = None):
        """
        Gets the lines that define an OpenShell3D geometry in a .geo file.

        :param update_data: Data used for VolumeModel defined with different shells
        :type update_data: dict
        :param point_mesh_size: The mesh size at a specific point, defaults to None
        :type point_mesh_size: float, optional
        :return: A list of lines that describe the geometry & the updated data
        :rtype: Tuple(List[str], dict)
        """

        primitives = []
        points = []

        for face in self.faces:
            for _, contour in enumerate(list(chain(*[[face.outer_contour3d], face.inner_contours3d]))):
                for point_contour in contour.get_geo_points():
                    if not point_in_list(point_contour, points):
                        points.append(point_contour)

                if isinstance(contour, curves.Circle2D):
                    pass
                else:
                    for _, primitive in enumerate(contour.primitives):
                        if (not edge_in_list(primitive, primitives)
                                and not edge_in_list(primitive.reverse(), primitives)):
                            primitives.append(primitive)

        indices_check = len(primitives) * [None]

        point_account = update_data['point_account']
        line_account, line_loop_account = update_data['line_account'] + 1, update_data['line_loop_account']
        lines, line_surface, lines_tags = [], [], []

        points = list(points)
        for p_index, point in enumerate(points):
            lines.append(point.get_geo_lines(tag=p_index + point_account + 1,
                                             point_mesh_size=point_mesh_size))

        for f_index, face in enumerate(self.faces):
            line_surface = []
            for _, contour in enumerate(list(chain(*[[face.outer_contour3d], face.inner_contours3d]))):
                lines_tags = []
                if isinstance(contour, curves.Circle2D):
                    pass
                else:
                    for _, primitive in enumerate(contour.primitives):
                        index = get_edge_index_in_list(primitive, primitives)

                        if primitives[index].is_close(primitive):

                            if isinstance(primitive, volmdlr.edges.BSplineCurve3D):
                                discretization_points = primitive.discretization_points()

                                start_point_tag = get_point_index_in_list(discretization_points[0], points) + 1
                                end_point_tag = get_point_index_in_list(discretization_points[1], points) + 1

                                primitive_linesegments = volmdlr.edges.LineSegment3D(
                                    discretization_points[0], discretization_points[1])
                                lines.append(primitive_linesegments.get_geo_lines(tag=line_account,
                                                                                  start_point_tag=start_point_tag
                                                                                  + point_account,
                                                                                  end_point_tag=end_point_tag
                                                                                  + point_account))

                            if isinstance(primitive, volmdlr.edges.LineSegment):

                                start_point_tag = get_point_index_in_list(primitive.start, points) + 1
                                end_point_tag = get_point_index_in_list(primitive.end, points) + 1

                                lines.append(primitive.get_geo_lines(tag=line_account,
                                                                     start_point_tag=start_point_tag + point_account,
                                                                     end_point_tag=end_point_tag + point_account))
                            elif isinstance(primitive, volmdlr.edges.ArcMixin):

                                start_point_tag = get_point_index_in_list(primitive.start, points) + 1
                                center_point_tag = get_point_index_in_list(primitive.circle.center, points) + 1
                                end_point_tag = get_point_index_in_list(primitive.end, points) + 1

                                lines.append(primitive.get_geo_lines(tag=line_account,
                                                                     start_point_tag=start_point_tag + point_account,
                                                                     center_point_tag=center_point_tag + point_account,
                                                                     end_point_tag=end_point_tag + point_account))

                            lines_tags.append(line_account)
                            indices_check[index] = line_account
                            line_account += 1

                        if primitives[index].is_close(primitive.reverse()):
                            lines_tags.append(-indices_check[index])

                    lines.append(contour.get_geo_lines(line_loop_account + 1, lines_tags))

                    line_surface.append(line_loop_account + 1)
                    line_loop_account += 1
                    lines_tags = []

            lines.append(face.get_geo_lines((f_index + 1 + update_data['surface_account']),
                                            line_surface))

            line_surface = []

        lines.append('Surface Loop(' + str(1 + update_data['surface_loop_account']) + ') = {'
                     + str(list(range(update_data['surface_account'] + 1,
                                      update_data['surface_account'] +
                                      len(self.faces) + 1)))[1:-1] + '};')

        update_data['point_account'] += len(points)
        update_data['line_account'] += line_account - 1
        update_data['line_loop_account'] += line_loop_account
        update_data['surface_account'] += len(self.faces)
        update_data['surface_loop_account'] += 1

        return lines, update_data

    def get_mesh_lines_with_transfinite_curves(self, min_points, size):
        """Gets Shells' mesh lines with transfinite curves."""
        lines = []
        for face in self.faces:
            lines.extend(face.surface2d.get_mesh_lines_with_transfinite_curves(
                [[face.outer_contour3d], face.inner_contours3d], min_points, size))
        return lines

    @staticmethod
    def is_shell_open(faces, faces_graph=None):
        """Returns True if shell is an open shell."""
        if faces_graph is None:
            vertices_points = Shell3D._helper_getter_vertices_points(faces)
            faces_graph = Shell3D._helper_create_faces_graph(faces, vertices_points, False)
        for n_index in faces_graph.nodes:
            face = faces[n_index]
            for prim in face.outer_contour3d.primitives:
                for neihgbor in faces_graph.neighbors(n_index):
                    if faces[neihgbor].outer_contour3d.is_primitive_section_over_wire(prim):
                        break
                else:
                    return True
        return False

    @classmethod
    def from_faces(cls, faces, name: str = ''):
        """
        Defines a List of separated OpenShell3D from a list of faces, based on the faces graph.
        """
        vertices_points = Shell3D._helper_getter_vertices_points(faces)
        graph = Shell3D._helper_create_faces_graph(faces, vertices_points, verify_connected_components=False)
        components = [graph.subgraph(c).copy() for c in nx.connected_components(graph)]

        shells_list = []
        for index, graph_i in enumerate(components, start=1):
            faces_list = [faces[n_index] for n_index in graph_i.nodes]
            if cls.is_shell_open(faces, graph_i):
                shells_list.append(OpenShell3D(faces_list, name=name + f'_{index}'))
            else:
                shells_list.append(ClosedShell3D(faces_list, name=name + f'_{index}'))

        return shells_list

    def is_disjoint_from(self, shell2, tol=1e-8):
        """
        Verifies and returns a Boolean if two shells are disjointed or not.

        """
        disjoint = True
        if self.bounding_box.is_intersecting(shell2.bounding_box, tol):
            return False
        return disjoint

    def is_face_intersecting(self, face: volmdlr.faces.Face3D):
        """Verifies if face is intersecting shell somehow."""
        if not self.bounding_box.is_intersecting(face.bounding_box):
            return False
        for i_face in self.faces:
            if i_face.face_intersections(face):
                return True
        return False

    def is_intersecting_with(self, shell2):
        """Verifies if two closed shells are intersecting somehow."""
        if self.is_disjoint_from(shell2):
            return False
        for face2 in shell2.faces:
            if self.is_face_intersecting(face2):
                return True
        return False


class OpenShell3D(Shell3D):
    """
    A 3D Open shell composed of multiple faces.

    This class represents a 3D open shell, which is a collection of connected
    faces with no volume. It is a subclass of the `Shell3D` class and
    inherits all of its attributes and methods.

    :param faces: The faces of the shell.
    :type faces: List[`Face3D`]
    :param color: The color of the shell.
    :type color: Tuple[float, float, float]
    :param alpha: The transparency of the shell, should be a value in the range (0, 1).
    :type alpha: float
    :param name: The name of the shell.
    :type name: str
    """

    STEP_FUNCTION = 'OPEN_SHELL'

    def union(self, shell2):
        """
        Combine two shells faces.

        :return: a new OpenShell3D with the combined faces.
        """
        new_faces = self.faces + shell2.faces
        new_name = self.name + ' union ' + shell2.name
        new_color = self.color
        return self.__class__(new_faces, name=new_name, color=new_color)


class ClosedShell3D(Shell3D):
    """
    A 3D closed shell composed of multiple faces.

    This class represents a 3D closed shell, which is a collection of connected
    faces with a volume. It is a subclass of the `Shell3D` class and
    inherits all of its attributes and methods. In addition, it has a method
    to check whether a face is inside the shell.

    :param faces: The faces of the shell.
    :type faces: List[`Face3D`]
    :param color: The color of the shell.
    :type color: Tuple[float, float, float]
    :param alpha: The transparency of the shell, should be a value in the range (0, 1).
    :type alpha: float
    :param name: The name of the shell.
    :type name: str
    """

    STEP_FUNCTION = 'CLOSED_SHELL'

    def volume(self):
        """
        Does not consider holes.

        """
        volume = 0
        center_x, center_y, center_z = self.bounding_box.center
        for face in self.faces:
            display3d = face.triangulation()
            for triangle_index in display3d.triangles:
                point1 = display3d.points[triangle_index[0]]
                point2 = display3d.points[triangle_index[1]]
                point3 = display3d.points[triangle_index[2]]

                point1_adj = (point1[0] - center_x, point1[1] - center_y, point1[2] - center_z)
                point2_adj = (point2[0] - center_x, point2[1] - center_y, point2[2] - center_z)
                point3_adj = (point3[0] - center_x, point3[1] - center_y, point3[2] - center_z)

                volume_tetraedre = 1 / 6 * abs(-point3_adj[0] * point2_adj[1] * point1_adj[2] +
                                               point2_adj[0] * point3_adj[1] * point1_adj[2] +
                                               point3_adj[0] * point1_adj[1] * point2_adj[2] -
                                               point1_adj[0] * point3_adj[1] * point2_adj[2] -
                                               point2_adj[0] * point1_adj[1] * point3_adj[2] +
                                               point1_adj[0] * point2_adj[1] * point3_adj[2])

                volume += volume_tetraedre

        return abs(volume)

    def is_face_inside(self, face: volmdlr.faces.Face3D):
        """
        Verifies if a face is inside the closed shell 3D.

        :param face: other face.
        :return: returns True if face is inside, and False otherwise.
        """
        if not face.bounding_box.is_inside_bbox(self.bounding_box):
            return False
        points = []
        if face.area() > 1e-8:
            points.append(face.random_point_inside())

        for prim in face.outer_contour3d.primitives:
            points.extend([prim.middle_point(), prim.end])
        for point in points:
            point_inside_shell = self.point_belongs(point)
            if not point_inside_shell:
                return False
        return True

    def get_ray_casting_line_segment(self, point3d):
        """Gets the best ray for performing ray casting algorithm."""
        boxes_size = [self.bounding_box.size[0] / 2, self.bounding_box.size[1] / 2, self.bounding_box.size[2] / 2]
        xyz = [volmdlr.Vector3D(boxes_size[0], 0, 0), volmdlr.Vector3D(0, boxes_size[1], 0),
               volmdlr.Vector3D(0, 0, boxes_size[2])]
        points = sorted(self.bounding_box.get_points_inside_bbox(2, 2, 2), key=point3d.point_distance)
        bbox_outside_points = []
        for vector in xyz:
            for direction in [1, -1]:
                bbox_outside_point = points[0] + direction * vector
                if not self.bounding_box.point_belongs(bbox_outside_point):
                    bbox_outside_points.append(bbox_outside_point)
        bbox_outside_points = sorted(bbox_outside_points, key=point3d.point_distance)
        vec1 = bbox_outside_points[0] - point3d
        vec1 = vec1.to_vector()
        vec2 = bbox_outside_points[1] - point3d
        vec2 = vec2.to_vector()
        vec3 = bbox_outside_points[2] - point3d
        vec3 = vec3.to_vector()
        rays = [edges.LineSegment3D(
                point3d, point3d + 2 * vec1 + random.random() * vec2 + random.random() * vec3) for _ in range(10)]
        return rays

    def point_belongs(self, point3d: volmdlr.Point3D, **kwargs):
        """
        Ray Casting algorithm.

        Returns True if the point is inside the Shell, False otherwise
        """
        bbox = self.bounding_box

        if not bbox.point_belongs(point3d):
            return False
        rays = self.get_ray_casting_line_segment(point3d)

        count = 0
        for ray in rays:
            count = 0
            intersections = []
            for _, point_inters in self.linesegment_intersections(ray):
                if point_inters[0].is_close(point3d):
                    return True
                for inter in point_inters:
                    if volmdlr.core.point_in_list(inter, intersections):
                        break
                    intersections.append(inter)
                    count += 1
                else:
                    continue
                break
            else:
                break
            continue
        is_inside = True
        if count % 2 == 0:
            is_inside = False
        return is_inside

    def is_inside_shell(self, shell2):
        """
        Returns True if all the points of self are inside shell2 and no face are intersecting.

        This method is not exact.
        """
        bbox1 = self.bounding_box
        bbox2 = shell2.bounding_box
        if not bbox1.is_inside_bbox(bbox2):
            return False
        for face in self.faces:
            if not shell2.is_face_inside(face):
                return False
        return True

    def intersecting_faces_combinations(self, shell2, tol=1e-8):
        """
        Gets intersecting faces combinations.

        :param shell2: ClosedShell3D
        :param tol: Corresponds to the tolerance to consider two faces as intersecting faces

        :return: returns a dictionary containing as keys the combination of intersecting faces
        and as the values the resulting primitive from the two intersecting faces.
        It is done so it is not needed to calculate the same intersecting primitive twice.
        """
        face_combinations = {}
        for face1 in self.faces:
            for face2 in shell2.faces:
                face_intersections = face1.face_intersections(face2, tol)
                if face_intersections:
                    face_combinations[(face1, face2)] = face_intersections
        return face_combinations

    @staticmethod
    def get_intersecting_faces(dict_intersecting_combinations):
        """
        Gets intersecting faces.

        :param dict_intersecting_combinations: dictionary containing as keys the combination of intersecting faces
        and as the values the resulting primitive from the two intersecting faces

        returns two lists. One for the intersecting faces in shell1 and the other for the shell2
        """
        intersecting_faces_shell1 = []
        intersecting_faces_shell2 = []
        for face in list(dict_intersecting_combinations.keys()):
            if face[0] not in intersecting_faces_shell1:
                intersecting_faces_shell1.append(face[0])
            if face[1] not in intersecting_faces_shell2:
                intersecting_faces_shell2.append(face[1])
        return intersecting_faces_shell1, intersecting_faces_shell2

    def get_non_intersecting_faces(self, shell2, intersecting_faces, intersection_method=False):
        """
        Gets lists of faces that never intersect with any of the shell2's faces.

        :param shell2: ClosedShell3D.
        :param intersecting_faces:
        :param intersection_method: determines if running for intersection operation.
        returns a list of all the faces that never intersect any
        face of the other shell.
        """
        non_intersecting_faces = [face for face in self.faces if face not in intersecting_faces]
        valid_non_intercting_faces = []
        for face in non_intersecting_faces:
            if shell2.face_on_shell(face):
                continue
            if intersection_method:
                if shell2.is_face_inside(face):
                    valid_non_intercting_faces.append(face)
            elif not shell2.is_face_inside(face):
                valid_non_intercting_faces.append(face)
        return valid_non_intercting_faces

    def get_coincident_faces(self, shell2):
        """
        Finds all pairs of faces that are coincident faces, that is, faces lying on the same plane.

        returns a List of tuples with the face pairs.
        """
        list_coincident_faces = []
        for face1 in self.faces:
            for face2 in shell2.faces:
                if isinstance(face1, face2.__class__) and face1.surface3d.is_coincident(face2.surface3d):
                    contour1 = face1.outer_contour3d.to_2d(
                        face1.surface3d.frame.origin,
                        face1.surface3d.frame.u,
                        face1.surface3d.frame.v)
                    contour2 = face2.outer_contour3d.to_2d(
                        face1.surface3d.frame.origin,
                        face1.surface3d.frame.u,
                        face1.surface3d.frame.v)
                    if contour1.bounding_rectangle.b_rectangle_intersection(contour2.bounding_rectangle):
                        list_coincident_faces.append((face1, face2))

        return list_coincident_faces

    def two_shells_intersecting_contour(self, shell2, dict_intersecting_combinations=None):
        """
        Computes intersecting_contour between two shells.

        :param shell2: ClosedShell3D
        :type shell2: :class:`volmdlr.faces.ClosedShell3D`.
        :param dict_intersecting_combinations: dictionary containing as keys
            the combination of intersecting faces and as the values the
            resulting primitive from the two intersecting faces
        :returns: intersecting contour for two intersecting shells
        """
        if dict_intersecting_combinations is None:
            dict_intersecting_combinations = self.intersecting_faces_combinations(shell2)
        intersecting_wires = list(dict_intersecting_combinations.values())
        intersecting_contour = \
            wires.Contour3D([wire.primitives[0] for
                             wires_ in intersecting_wires for wire in wires_])
        return intersecting_contour

    def reference_shell(self, shell2, face):
        """Reference shell used during bool operations, to help decide if a new divided face should be saved or not."""
        if face in shell2.faces:
            reference_shell = self
        else:
            reference_shell = shell2
        return reference_shell

    def set_operations_valid_exterior_faces(self, new_faces: List[volmdlr.faces.Face3D],
                                            valid_faces: List[volmdlr.faces.Face3D],
                                            list_coincident_faces: List[volmdlr.faces.Face3D],
                                            shell2, reference_shell):
        """
        Select the valid faces from the new faces created during Boolean operations.

        :param new_faces: list of new divided faces.
        :param valid_faces: list of already validated faces.
        :param list_coincident_faces: if of coincident faces.
        :param shell2: shell2, used in the Boolean operation.
        :param reference_shell: reference shell, to help decide if a new divided face should be saved or not.
        :return:
        """
        for new_face in new_faces:
            if self.set_operations_exterior_face(new_face, valid_faces, reference_shell,
                                                 list_coincident_faces, shell2):
                valid_faces.append(new_face)
        return valid_faces

    def union_faces(self, shell2, intersecting_faces, intersecting_combinations, list_coincident_faces):
        """
        Gets new faces for union Boolean operation between two closed shell 3d.

        :param shell2: other shell
        :param intersecting_faces: list of all intersecting faces.
        :param intersecting_combinations: Dictionary containing all combination of faces intersection,\
        with corresponding intersections.
        :param list_coincident_faces: list of coincident faces.
        :return: list of new faces for union of two closed shell3.
        """
        faces = []
        for face in intersecting_faces:
            reference_shell = self.reference_shell(shell2, face)
            new_faces = face.set_operations_new_faces(intersecting_combinations)
            faces = self.set_operations_valid_exterior_faces(new_faces, faces, list_coincident_faces,
                                                             shell2, reference_shell)
        if list_coincident_faces:
            faces = self.validate_set_operations_faces(faces)
        return faces

    def get_subtraction_valid_faces(self, new_faces, valid_faces, reference_shell, shell2, keep_interior_faces):
        """
        Gets valid faces for subtraction Boolean operations.

        :param new_faces: list of new divided faces.
        :param valid_faces: list of already validated faces.
        :param reference_shell: reference shell, to help decide if a new divided face should be saved or not.
        :param shell2: other shell.
        :param keep_interior_faces: Boolean to decide to keep interior faces on reference shell or not.
        :return: return a list a valid faces.
        """
        faces = []
        for new_face in new_faces:
            if reference_shell.face_on_shell(new_face):
                if (reference_shell == self and shell2.face_on_shell(new_face)) or (
                        reference_shell == shell2 and self.face_on_shell(new_face)):
                    if self.is_face_between_shells(shell2, new_face):
                        faces.append(new_face)
                continue
            if keep_interior_faces:
                if self.set_operations_interior_face(new_face, valid_faces, reference_shell):
                    faces.append(new_face)
            elif self.set_operations_exterior_face(new_face, faces, reference_shell, [], shell2):
                faces.append(new_face)
        return faces

    @staticmethod
    def validate_set_operations_faces(faces):
        """
        Final validation of new faces created during intersections or subtractions of two closed shells.

        :param faces: new faces.
        :return: valid faces.
        """
        valid_faces = []
        while True:
            if not faces:
                break
            for face in valid_faces:
                if face.face_inside(faces[0]):
                    faces.remove(faces[0])
                    break
            else:
                valid_faces.append(faces[0])
                faces.remove(faces[0])
        return valid_faces

    def subtraction_faces(self, shell2, intersecting_faces, intersecting_combinations):
        """
        Gets new faces for subtraction Boolean operation between two closed shell 3d.

        :param shell2: other shell
        :param intersecting_faces: list of all intersecting faces.
        :param intersecting_combinations: Dictionary containing all combination of faces intersection,\
        with corresponding intersections.
        :return: list of new faces for subtraction of two closed shells 3.
        """
        faces = []
        for face in intersecting_faces:
            keep_interior_faces = False
            if face in shell2.faces:
                keep_interior_faces = True
            reference_shell = self.reference_shell(shell2, face)
            new_faces = face.set_operations_new_faces(intersecting_combinations)
            valid_faces = self.get_subtraction_valid_faces(new_faces, faces, reference_shell,
                                                           shell2, keep_interior_faces)
            faces.extend(valid_faces)

        valid_faces = self.validate_set_operations_faces(faces)

        return valid_faces

    def valid_intersection_faces(self, new_faces, valid_faces, reference_shell, shell2):
        """
        Validate Boolean intersection operation new faces.

        :param new_faces: list of new divided faces.
        :param valid_faces: list of already validated faces.
        :param reference_shell: reference shell, to help decide if a new divided face should be saved or not.
        :param shell2: other shell.
        :return:
        """
        faces = []
        for new_face in new_faces:
            if reference_shell.face_on_shell(new_face):
                if (reference_shell == self and shell2.face_on_shell(new_face)) or (
                        reference_shell == shell2 and self.face_on_shell(new_face)):
                    point3d = new_face.random_point_inside()
                    if new_face.point_belongs(point3d):
                        normal1 = point3d - 0.00001 * new_face.surface3d.frame.w.unit_vector()
                        normal2 = point3d + 0.00001 * new_face.surface3d.frame.w.unit_vector()
                        if (self.point_belongs(normal1) and shell2.point_belongs(normal1)) or \
                                (shell2.point_belongs(normal2) and self.point_belongs(normal2)):
                            faces.append(new_face)
                    continue
            inside_reference_shell = reference_shell.point_belongs(
                new_face.random_point_inside())
            if inside_reference_shell and new_face not in valid_faces:
                faces.append(new_face)

        return faces

    def intersection_faces(self, shell2, intersecting_faces, intersecting_combinations):
        """
        Gets new faces for intersection Boolean operation between two closed shell 3d.

        :param shell2: other shell
        :param intersecting_faces: list of all intersecting faces.
        :param intersecting_combinations: Dictionary containing all combination of faces intersection,\
        with corresponding intersections.
        :return: list of new faces for intersection of two closed shells 3d.
        """
        faces = []
        for face in intersecting_faces:
            reference_shell = self.reference_shell(shell2, face)
            new_faces = face.set_operations_new_faces(intersecting_combinations)
            valid_faces = self.valid_intersection_faces(
                new_faces, faces, reference_shell, shell2)
            faces.extend(valid_faces)

        valid_faces = self.validate_set_operations_faces(faces)
        return valid_faces

    def set_operations_interior_face(self, new_face, faces, reference_shell):
        """
        Verify if new face is inside reference shell for Boolean operations.

        :param new_face: new divided face.
        :param faces: list of already validated faces.
        :param reference_shell: reference shell, to help decide if a new divided face should be saved or not.
        """
        inside_reference_shell = reference_shell.point_belongs(new_face.random_point_inside())
        if inside_reference_shell and new_face not in faces:
            return True
        if self.face_on_shell(new_face):
            return True
        return False

    def is_face_between_shells(self, shell2, face):
        """
        Verify if face is between the two shells.

        :param shell2: other shell
        :param face: face to be verified.
        :return:
        """
        if face.surface2d.inner_contours:
            normal_0 = face.surface2d.outer_contour.primitives[0].normal_vector()
            middle_point_0 = face.surface2d.outer_contour.primitives[0].middle_point()
            point1 = middle_point_0 + 0.0001 * normal_0
            point2 = middle_point_0 - 0.0001 * normal_0
            points = [point1, point2]
        else:
            points = [face.surface2d.outer_contour.random_point_inside()]

        for point in points:
            point3d = face.surface3d.point2d_to_3d(point)
            if face.point_belongs(point3d):
                normal1 = point3d - 0.00001 * face.surface3d.frame.w
                normal2 = point3d + 0.00001 * face.surface3d.frame.w
                if (self.point_belongs(normal1) and
                    shell2.point_belongs(normal2)) or \
                        (shell2.point_belongs(normal1) and
                         self.point_belongs(normal2)):
                    return True
        return False

    def set_operations_exterior_face(self, new_face, valid_faces, reference_shell,
                                     list_coincident_faces, shell2):
        """
        Selects exterior faces during bool operations, like union or subtraction.

        :param new_face: divided faces.
        :param valid_faces: list of already validated faces.
        :param reference_shell: reference shell, to help decide if a new divided face should be saved or not.
        :param shell2: other shell.
        :param list_coincident_faces: list of coincident faces.
        :return:
        """
        if new_face.area() < 1e-8:
            return False
        if new_face not in valid_faces:
            inside_reference_shell = reference_shell.point_belongs(new_face.random_point_inside())
            face_on_reference_shell = reference_shell.face_on_shell(new_face)
            if not inside_reference_shell or face_on_reference_shell:
                if list_coincident_faces:
                    if self.is_face_between_shells(shell2, new_face):
                        return False
                return True
        return False

    def validate_set_operation(self, shell2, tol):
        """
        Verifies if two shells are valid for union or subtractions operations.

        Its Verifies if they are disjointed or if one is totally inside the other.

        If it returns an empty list, it means the two shells are valid to continue the
        operation.
        """
        if self.is_disjoint_from(shell2, tol):
            return [self, shell2]
        if self.is_inside_shell(shell2):
            return [shell2]
        if shell2.is_inside_shell(self):
            return [self]
        return []

    def union(self, shell2: 'ClosedShell3D', tol: float = 1e-8):
        """
        Given Two closed shells, it returns a new united ClosedShell3D object.

        """
        validate_set_operation = self.validate_set_operation(shell2, tol)
        if validate_set_operation:
            return validate_set_operation
        list_coincident_faces = self.get_coincident_faces(shell2)
        intersecting_combinations = self.intersecting_faces_combinations(shell2, tol)
        intersecting_faces1, intersecting_faces2 = self.get_intersecting_faces(intersecting_combinations)
        intersecting_faces = intersecting_faces1 + intersecting_faces2
        faces = self.get_non_intersecting_faces(shell2, intersecting_faces)
        faces += shell2.get_non_intersecting_faces(self, intersecting_faces)
        if len(faces) == len(self.faces + shell2.faces) and not intersecting_faces:
            return [self, shell2]
        new_valid_faces = self.union_faces(shell2, intersecting_faces,
                                           intersecting_combinations, list_coincident_faces)
        faces += new_valid_faces
        new_shell = ClosedShell3D(faces)
        return [new_shell]

    @staticmethod
    def get_faces_to_be_merged(union_faces):
        """Gets faces that are adjacent, and sharing the same surface, so they can be merged."""
        coincident_planes_faces = []
        for i, face1 in enumerate(union_faces):
            for j, face2 in enumerate(union_faces):
                if j != i and face1.surface3d.is_coincident(face2.surface3d):
                    if face1 not in coincident_planes_faces:
                        coincident_planes_faces.append(face1)
                    coincident_planes_faces.append(face2)
            if coincident_planes_faces:
                break
        return coincident_planes_faces

    @staticmethod
    def clean_faces(union_faces, list_new_faces):
        list_remove_faces = []
        if union_faces:
            for face1 in union_faces:
                for face2 in list_new_faces:
                    if face1.face_inside(face2):
                        list_remove_faces.append(face2)
                    elif face2.face_inside(face1):
                        list_remove_faces.append(face1)
        list_new_faces += union_faces
        for face in list_remove_faces:
            list_new_faces.remove(face)
        return list_new_faces

    def merge_faces(self):
        """
        Merges all shells' adjacent faces into one.

        """
        union_faces = self.faces
        finished = False
        list_new_faces = []
        count = 0
        while not finished:
            valid_coicident_faces = ClosedShell3D.get_faces_to_be_merged(union_faces)
            list_valid_coincident_faces = valid_coicident_faces[:]
            if valid_coicident_faces:
                list_new_faces += volmdlr.faces.PlaneFace3D.merge_faces(valid_coicident_faces)
            for face in list_valid_coincident_faces:
                union_faces.remove(face)
            count += 1
            if count >= len(self.faces) and not list_valid_coincident_faces:
                finished = True

        list_new_faces = self.clean_faces(union_faces, list_new_faces)

        self.faces = list_new_faces

    def subtract(self, shell2, tol=1e-8):
        """
        Given Two closed shells, it returns a new subtracted OpenShell3D.

        """
        validate_set_operation = self.validate_set_operation(shell2, tol)
        if validate_set_operation:
            return validate_set_operation

        list_coincident_faces = self.get_coincident_faces(shell2)
        intersecting_combinations = self.intersecting_faces_combinations(shell2, tol)

        if len(intersecting_combinations) == 0:
            return [self, shell2]

        intersecting_faces, _ = self.get_intersecting_faces(
            intersecting_combinations)

        faces = self.get_non_intersecting_faces(shell2, intersecting_faces)
        new_valid_faces = self.union_faces(shell2, intersecting_faces,
                                           intersecting_combinations,
                                           list_coincident_faces
                                           )
        faces += new_valid_faces
        return OpenShell3D.from_faces(faces)

    def subtract_to_closed_shell(self, shell2: OpenShell3D, tol: float = 1e-8):
        """
        Subtracts shell2's volume from self.

        :param shell2: other shell
        :param tol: tolerance
        :return:
        """
        if self.is_disjoint_from(shell2, tol):
            return [self]
        intersecting_combinations = self.intersecting_faces_combinations(shell2, tol)

        intersecting_faces1, intersecting_faces2 = self.get_intersecting_faces(
            intersecting_combinations)
        intersecting_faces = intersecting_faces1 + intersecting_faces2

        faces = self.get_non_intersecting_faces(shell2, intersecting_faces)
        faces += shell2.get_non_intersecting_faces(self, intersecting_faces, intersection_method=True)
        new_valid_faces = self.subtraction_faces(shell2, intersecting_faces, intersecting_combinations)
        faces += new_valid_faces
        # new_shell = ClosedShell3D(faces)
        # new_shell.eliminate_not_valid_closedshell_faces()
        return self.from_faces(faces)

    def validate_intersection_operation(self, shell2):
        """
        Verifies if two shells are valid for union or subtractions operations.

        Its Verifies if they are disjointed or if one is totally inside the other.
        If it returns an empty list, it means the two shells are valid to continue the
        operation.
        """
        if self.is_inside_shell(shell2):
            return [self]
        if shell2.is_inside_shell(self):
            return [shell2]
        return []

    def intersection(self, shell2, tol=1e-8):
        """
        Given two ClosedShell3D, it returns the new object resulting from the intersection of the two.

        """
        if self.is_disjoint_from(shell2, tol):
            return []
        validate_set_operation = self.validate_intersection_operation(shell2)
        if validate_set_operation:
            return validate_set_operation
        intersecting_combinations = self.intersecting_faces_combinations(shell2, tol)

        if not intersecting_combinations:
            return [self, shell2]

        intersecting_faces1, intersecting_faces2 = self.get_intersecting_faces(intersecting_combinations)
        intersecting_faces = intersecting_faces1 + intersecting_faces2
        faces = self.intersection_faces(shell2, intersecting_faces, intersecting_combinations)
        faces += self.get_non_intersecting_faces(shell2, intersecting_faces, intersection_method=True)
        faces += shell2.get_non_intersecting_faces(self, intersecting_faces, intersection_method=True)
        new_shell = ClosedShell3D(faces)
        return [new_shell]

    def eliminate_not_valid_closedshell_faces(self):
        nodes_with_2degrees = [node for node, degree in list(self.vertices_graph.degree()) if degree <= 2]
        for node in nodes_with_2degrees:
            neighbors = nx.neighbors(self.vertices_graph, node)
            for neighbor_node in neighbors:
                for face in self.faces:
                    if self.vertices_graph.edges[(node, neighbor_node)]['edge'] in face.outer_contour3d.primitives:
                        self.faces.remove(face)
                        break
        self._faces_graph = None


class OpenTriangleShell3D(OpenShell3D):
    """
    A 3D open shell composed of multiple triangle faces.

    This class represents a 3D open shell, which is a collection of connected
    triangle faces with no volume. It is a subclass of the `OpenShell3D` class
    and inherits all of its attributes and methods.

    :param faces: The triangle faces of the shell.
    :type faces: List[`Triangle3D`]
    :param color: The color of the shell.
    :type color: Tuple[float, float, float]
    :param alpha: The transparency of the shell, should be a value in the range (0, 1).
    :type alpha: float
    :param name: The name of the shell.
    :type name: str
    """

    def __init__(
        self,
        faces: List[volmdlr.faces.Triangle3D],
        color: Tuple[float, float, float] = None,
        alpha: float = 1.0,
        name: str = "",
    ):
        OpenShell3D.__init__(self, faces=faces, color=color, alpha=alpha, name=name)

    def get_bounding_box(self) -> volmdlr.core.BoundingBox:
        """Gets the Shell bounding box."""
        vertices = np.array(
            [(face.points[i].x, face.points[i].y, face.points[i].z) for face in self.faces for i in range(3)]
        )
        bbox_min, bbox_max = np.min(vertices, axis=0), np.max(vertices, axis=0)

        return volmdlr.core.BoundingBox(bbox_min[0], bbox_max[0], bbox_min[1], bbox_max[1], bbox_min[2], bbox_max[2])

    def to_mesh_data(self, round_vertices: bool, n_decimals: int = 9) -> Tuple[NDArray[float], NDArray[int]]:
        """
        Convert the TriangleShell3D to mesh data: vertices and faces described as index of vertices.

        :param round_vertices: Allows to choose to round vertices coordinates or not.
            Rounding vertices coordinates allows to prevent numerical imprecision, which allows vertex sharing between
            adjacent triangles.
        :type round_vertices: bool
        :param n_decimals: int
        :type n_decimals: float

        :return: The vertices and faces composing the mesh data.
        :rtype: Tuple[NDArray[float], NDArray[int]]
        """
        # Flatten and round the vertices array
        vertices = np.array(
            [(face.points[i].x, face.points[i].y, face.points[i].z) for face in self.faces for i in range(3)]
        )

        if round_vertices:
            vertices = np.round(vertices, n_decimals)  # rounding to prevent numerical imprecision

        # Get unique vertices and their indices
        vertices, unique_indices = np.unique(vertices, axis=0, return_inverse=True)

        # Create the triangle indices array using NumPy indexing
        flattened_indices = unique_indices.reshape(-1, 3)
        faces = flattened_indices[: len(self.faces)]

        return vertices, faces

    @classmethod
    def from_mesh_data(cls, vertices: Iterable[Iterable[float]], faces: Iterable[Iterable[int]], name: str = ""):
        """
        Create a TriangleShell3D from mesh data: vertices and faces described as index of vertices.

        :param vertices: The vertices of the mesh.
        :type vertices: Iterable[Iterable[float]]
        :param faces: The faces of the mesh, using vertices indexes.
        :type faces: Iterable[Iterable[int]]
        :param name: A name for the TriangleShell3D, optional.
        :type name: str

        :return: The created TriangleShell3D.
        :rtype: TriangleShell3D
        """
        triangles = []

        points = [volmdlr.Point3D(px, py, pz) for px, py, pz in vertices]

        for i1, i2, i3 in faces:
            triangles.append(volmdlr.faces.Triangle3D(points[i1], points[i2], points[i3]))

        return cls(triangles, name=name)

    def to_trimesh(self):
        """Creates a Trimesh from a TriangleShell3D."""
        return Trimesh(*self.to_mesh_data(round_vertices=True))

    @classmethod
    def from_trimesh(cls, trimesh, name: str = ""):
        """Creates a TriangleShell3D from Trimesh."""
        return cls.from_mesh_data(trimesh.vertices, trimesh.faces, name=name)

    def triangulation(self):
        """Triangulation of an Open Triangle Shell 3D."""
        points = []
        triangles = []
        for i, triangle in enumerate(self.faces):
            points.append(display.Node3D.from_point(triangle.point1))
            points.append(display.Node3D.from_point(triangle.point2))
            points.append(display.Node3D.from_point(triangle.point3))
            triangles.append((3 * i, 3 * i + 1, 3 * i + 2))

        return display.DisplayMesh3D(points, triangles)

    def to_dict(self, *args, **kwargs):
        """Overload of 'to_dict' for performance."""
        dict_ = self.base_dict()

        # not rounding to make sure to retrieve the exact same object with 'dict_to_object'
        vertices, faces = self.to_mesh_data(round_vertices=False)

        dict_["vertices"] = vertices.tolist()
        dict_["faces"] = faces.tolist()
        dict_["alpha"] = self.alpha
        dict_["color"] = self.color

        return dict_

    @classmethod
    def dict_to_object(
        cls,
        dict_: JsonSerializable,
        force_generic: bool = False,
        global_dict=None,
        pointers_memo: Dict[str, Any] = None,
        path: str = "#",
        name: str = "",
    ) -> "OpenTriangleShell3D":
        """Overload of 'dict_to_object' for performance."""

        vertices = dict_["vertices"]
        faces = dict_["faces"]
        name = dict_["name"]

        triangle_shell = cls.from_mesh_data(vertices, faces, name)
        triangle_shell.alpha = dict_["alpha"]
        triangle_shell.color = dict_["color"]

        return triangle_shell

    def to_display_triangle_shell(self) -> "DisplayTriangleShell3D":
        """
        Create a DisplayTriangleShell3D from the current TriangleShell3D.

        :return: The created DisplayTriangleShell3D.
        :rtype: DisplayTriangleShell3D
        """
        return DisplayTriangleShell3D.from_triangle_shell(self)


class ClosedTriangleShell3D(OpenTriangleShell3D, ClosedShell3D):
    """
    A 3D closed shell composed of multiple triangle faces.

    This class represents a 3D closed shell, which is a collection of connected
    triangle faces with a volume. It is a subclass of both the `ClosedShell3D`
    and `OpenTriangleShell3D` classes and inherits all of their attributes and
    methods.

    :param faces: The triangle faces of the shell.
    :type faces: List[`Triangle3D`]
    :param color: The color of the shell.
    :type color: Tuple[float, float, float]
    :param alpha: The transparency of the shell, should be a value in the range (0, 1).
    :type alpha: float
    :param name: The name of the shell.
    :type name: str
    """

    def __init__(
        self,
        faces: List[volmdlr.faces.Triangle3D],
        color: Tuple[float, float, float] = None,
        alpha: float = 1.0,
        name: str = "",
    ):
        OpenTriangleShell3D.__init__(self, faces=faces, color=color, alpha=alpha, name=name)
        ClosedShell3D.__init__(self, faces, color, alpha, name)


class DisplayTriangleShell3D(Shell3D):
    """
    A Triangle Shell 3D optimized for display and saving purpose.

    This shell has the particularity to not instantiate the Triangle3D objects, to reduce memory usage and improve
    performance.
    """

    def __init__(self, positions: NDArray[float], indices: NDArray[int], name):
        """
        Instantiate the DisplayTriangleShell3D.

        :param positions: A 3D numpy array of float representing the positions of the vertices of the triangles.
        :param indices: A 3D numpy array of int representing the indices of the vertices representing the triangles.
        :param name: A name for the DisplayTriangleShell3D, optional.
        """
        self.positions = positions
        self.indices = indices

        Shell3D.__init__(self, faces=[], name=name)  # avoid saving the faces for memory and performance

    @classmethod
    def from_triangle_shell(
        cls, triangle_shell: Union["OpenTriangleShell3D", "ClosedTriangleShell3D"]
    ) -> "DisplayTriangleShell3D":
        """
        Instantiate a DisplayTriangleShell3D from an OpenTriangleShell3D or a ClosedTriangleShell3D.

        :param triangle_shell: The triangle shell to create the DisplayTriangleShell3D from.
        :type triangle_shell: OpenTriangleShell3D | ClosedTriangleShell3D

        :return: The created DisplayTriangleShell3D.
        :rtype: DisplayTriangleShell3D
        """
        positions, indices = triangle_shell.to_mesh_data(round_vertices=True, n_decimals=6)
        name = triangle_shell.name

        display_triangle_shell = cls(positions, indices, name)

        display_triangle_shell.alpha = triangle_shell.alpha
        display_triangle_shell.color = triangle_shell.color

        return display_triangle_shell

    def get_bounding_box(self) -> volmdlr.core.BoundingBox:
        """Gets the Shell bounding box."""
        bbox_min, bbox_max = np.min(self.positions, axis=0), np.max(self.positions, axis=0)

        return volmdlr.core.BoundingBox(bbox_min[0], bbox_max[0], bbox_min[1], bbox_max[1], bbox_min[2], bbox_max[2])

    def babylon_meshes(self, merge_meshes=True):
        """Overload of 'babylon_meshes' for performance."""

        babylon_mesh = {"positions": self.positions.flatten().tolist(), "indices": self.indices.flatten().tolist()}
        babylon_mesh.update(self.babylon_param())

        return [babylon_mesh]

    def to_dict(self, *args, **kwargs):
        """Overload of 'to_dict' for performance."""
        dict_ = self.base_dict()

        dict_["positions"] = self.positions.tolist()
        dict_["indices"] = self.indices.tolist()
        dict_["alpha"] = self.alpha
        dict_["color"] = self.color

        return dict_

    @classmethod
    def dict_to_object(
        cls,
        dict_: JsonSerializable,
        force_generic: bool = False,
        global_dict=None,
        pointers_memo: Dict[str, Any] = None,
        path: str = "#",
        name: str = "",
    ) -> "DisplayTriangleShell3D":
        """Overload of 'dict_to_object' for performance."""

        positions = np.array(dict_["positions"])
        indices = np.array(dict_["indices"])
        name = dict_["name"]

        display_triangle_shell = cls(positions, indices, name)

        display_triangle_shell.alpha = dict_["alpha"]
        display_triangle_shell.color = dict_["color"]

        return display_triangle_shell



"""some volmdlr edges models."""
from geomdl import utilities
import volmdlr
from volmdlr import edges, curves

DEGREE = 3
points = [volmdlr.Point2D(0, 0), volmdlr.Point2D(1, 1), volmdlr.Point2D(2, -1), volmdlr.Point2D(3, 0)]
knotvector = utilities.generate_knot_vector(DEGREE, len(points))
knot_multiplicity = [1] * len(knotvector)
bspline1 = edges.BSplineCurve2D(DEGREE, points, knot_multiplicity, knotvector, None)

lineseg = edges.LineSegment2D(volmdlr.Point2D(0, 0.2), volmdlr.Point2D(3, -0.2))

arc = edges.Arc2D.from_3_points(volmdlr.Point2D(0, 0.3), volmdlr.Point2D(1, -0.3), volmdlr.Point2D(2, 2))
u_vector = volmdlr.Vector2D(0.7071067811865475, 0.7071067811865475)
v_vector = volmdlr.Vector2D(-0.7071067811865475, 0.7071067811865475)
frame = volmdlr.Frame2D(volmdlr.O2D, u_vector, v_vector)
ellipse2d = curves.Ellipse2D(2, 1, frame)
arc_ellipse2d = edges.ArcEllipse2D(ellipse2d, start=volmdlr.Point2D(0.5, 1.5), end=volmdlr.Point2D(1.5, 0.5))


def bspline_curve3d():
    """Bspline curve3d model."""
    degree = 5
    control_points = [volmdlr.Point3D(0, 3, 0),
                      volmdlr.Point3D(3, 2, 1),
                      volmdlr.Point3D(5, -1, 4),
                      volmdlr.Point3D(5, -4, 0),
                      volmdlr.Point3D(-1, -2, -3),
                      volmdlr.Point3D(-3, 4, 1)]
    knots = [0.0, 1.0]
    knot_multiplicities = [6, 6]
    weights = None  # [1, 2, 1, 2, 1, 2]
    return edges.BSplineCurve3D(degree=degree, control_points=control_points,
                                knot_multiplicities=knot_multiplicities,
                                knots=knots,
                                weights=weights,
                                name='B Spline Curve 3D 1')


vector1 = volmdlr.Vector3D(1, 1, 1)
vector1 = vector1.unit_vector()
vector2 = vector1.deterministic_unit_normal_vector()
vector3 = vector1.cross(vector2)


def arc3d():
    """Arc 3d model."""
    circle3d = curves.Circle3D(volmdlr.Frame3D(volmdlr.O3D, vector1, vector2, vector3), 1)
    return edges.Arc3D(circle3d, start=volmdlr.Point3D(0.5773502691896258, 0.5773502691896258, 0.5773502691896258),
                       end=volmdlr.Point3D(-0.9855985596534886, -0.11957315586905026, -0.11957315586905026))


def arc_ellipse3d():
    """Arc ellipse 3d model."""
    ellipse3d = curves.Ellipse3D(2, 1, volmdlr.Frame3D(volmdlr.Point3D(1, 2, 1), vector3, vector1, vector2))
    return edges.ArcEllipse3D(ellipse3d,
                              start=volmdlr.Point3D(0.42264973081037405, 1.4226497308103743, 0.42264973081037427),
                              end=volmdlr.Point3D(1.577350269189626, 2.5773502691896253, 1.5773502691896257))


def linesegment3d():
    """Linesegment 3d model."""
    return edges.LineSegment3D(volmdlr.Point3D(1, 2, 4), volmdlr.Point3D(-1, 5, -3))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=invalid-unary-operand-type
"""
Module containing mesh and relative objects.
"""

import math
from itertools import combinations
from typing import List

import matplotlib.pyplot as plt
import numpy as npy

from dessia_common.core import DessiaObject  # isort: skip

import volmdlr as vm
import volmdlr.edges as vme
import volmdlr.gmsh_vm
import volmdlr.wires as vmw
from volmdlr.core import EdgeStyle


class FlatElementError(Exception):
    """An error in case an element is flat."""


# def find_duplicate_linear_element(linear_elements1, linear_elements2):
#     duplicates = []
#     for linear_element in linear_elements1:
#         if linear_element in linear_elements2 and linear_element not in duplicates:
#             duplicates.append(linear_element)
#     return duplicates


class Node2D(vm.Point2D):
    """
    A node is a Point2D with some hash capabilities for performance used for Mesh.

    """

    def __init__(self, x: float, y: float, name: str = ""):
        self.x = x
        self.y = y
        vm.Point2D.__init__(self, x, y, name)

    def __hash__(self):
        return int(1e6 * (self.x + self.y))

    def __eq__(self, other_node: 'Node2D'):
        if other_node.__class__.__name__ not in ['Vector2D', 'Point2D',
                                                 'Node2D']:
            return False
        return math.isclose(self.x, other_node.x, abs_tol=1e-12) \
            and math.isclose(self.y, other_node.y, abs_tol=1e-12)

    @classmethod
    def from_point(cls, point2d, name: str = ''):
        """
        Defines a node2d from a point2d.

        :param point2d: A point2d
        :type point2d: vm.Point2D.
        :param name: object's name.
        :return: A node2d
        :rtype: Node2D
        """

        return cls(point2d.x, point2d.y, name=name)


class Node3D(vm.Point3D):
    """
    A node is a Point3D with some hash capabilities for performance used for Mesh.
    """

    def __init__(self, x: float, y: float, z: float, name: str = ""):
        self.x = x
        self.y = y
        self.z = z
        vm.Point3D.__init__(self, x, y, z, name)

    def __hash__(self):
        return int(1e6 * (self.x + self.y + self.z))

    def __eq__(self, other_node: 'Node3D'):
        if other_node.__class__.__name__ not in ['Vector3D', 'Point3D',
                                                 'Node3D']:
            return False
        return math.isclose(self.x, other_node.x, abs_tol=1e-12) \
            and math.isclose(self.y, other_node.y, abs_tol=1e-12)

    @classmethod
    def from_point(cls, point3d, name: str = ''):
        """
        Defines a node3d from a point3d.

        :param point3d: A point3d
        :type point3d: vm.Point3D.
        :param name: object's name.
        :return: A node3d
        :rtype: Node3D
        """

        return cls(point3d.x, point3d.y, point3d.z, name=name)


class LinearElement(vme.LineSegment2D):
    """ A class that defines a linear element. """
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, start: vm.Point2D, end: vm.Point2D,
                 interior_normal: vm.Vector2D, name: str = ''):
        self.points = [start, end]
        self.interior_normal = interior_normal

        vme.LineSegment2D.__init__(self, start=start, end=end, name=name)

    # def __hash__(self):
    #     return self.start.__hash__() + self.end.__hash__()

    # def __eq__(self, other_linear_element):
    #     if self.__class__ != other_linear_element.__class__:
    #         return False
    #     return (self.start == other_linear_element.start and self.end == other_linear_element.end) \
    #         or (self.start== other_linear_element.end and self.end == other_linear_element.start)

    # def plot(self, ax=None, color='k', width=None, plot_points=False):
    #     if ax is None:
    #         fig, ax = plt.subplots()
    #         ax.set_aspect('equal')
    #     if width is None:
    #         width=1
    #     if plot_points:
    #         ax.plot([self.start.x, self.end.x], [self.start.y, self.end.y], color=color, marker='o', linewidth=width)
    #     else:
    #         ax.plot([self.start.x, self.end.x], [self.start.y, self.end.y], color=color, linewidth=width)
    #     return ax


class TriangularElement(vmw.Triangle):
    """
    A mesh element defined with 3 nodes.
    """
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, points: List[volmdlr.Point2D]):
        # super().__init__(*points)
        self.points = points
        # self.linear_elements = self._to_linear_elements()
        # self.form_functions = self._form_functions()
        # self.line_segments = self.line_segments()
        self.center = (self.points[0] + self.points[1] + self.points[2]) / 3

        self._line_segments = None

        # self.area = self._area()

        vmw.Triangle.__init__(self, *points)

    def _to_linear_elements(self):
        vec1 = vm.Vector2D(self.points[1].x - self.points[0].x,
                           self.points[1].y - self.points[0].y)
        vec2 = vm.Vector2D(self.points[2].x - self.points[1].x,
                           self.points[2].y - self.points[1].y)
        vec3 = vm.Vector2D(self.points[0].x - self.points[2].x,
                           self.points[0].y - self.points[2].y)
        normal1 = vm.Vector2D(-vec1.y, vec1.x)
        normal2 = vm.Vector2D(-vec2.y, vec2.x)
        normal3 = vm.Vector2D(-vec3.y, vec3.x)
        normal1 = normal1.unit_vector()
        normal2 = normal2.unit_vector()
        normal3 = normal3.unit_vector()
        if normal1.dot(vec2) < 0:
            normal1 = - normal1
        if normal2.dot(vec3) < 0:
            normal2 = - normal2
        if normal3.dot(vec1) < 0:
            normal3 = - normal3
        linear_element_1 = LinearElement(self.points[0], self.points[1],
                                         normal1)
        linear_element_2 = LinearElement(self.points[1], self.points[2],
                                         normal2)
        linear_element_3 = LinearElement(self.points[2], self.points[0],
                                         normal3)
        return [linear_element_1, linear_element_2, linear_element_3]

    def _form_functions(self):
        a_matrix = vm.Matrix33(1, self.points[0].x, self.points[0].y,
                               1, self.points[1].x, self.points[1].y,
                               1, self.points[2].x, self.points[2].y)
        try:
            inv_a = a_matrix.inverse()
        except ValueError as expt:
            # self.plot()
            print('buggy element area', self._area())
            raise FlatElementError('form function bug') from expt
        x_1 = inv_a.vector_multiplication(vm.X3D)
        x_2 = inv_a.vector_multiplication(vm.Y3D)
        x_3 = inv_a.vector_multiplication(vm.Z3D)
        return x_1, x_2, x_3

    # def _quadratic_form_functions(self):
    #     a = [[1, self.points[0][0], self.points[0][1],self.points[0][0]**2,
    #           self.points[0][0]*self.points[0][1],self.points[0][1]**2],
    #           [1, self.points[1][0], self.points[1][1],self.points[1][0]**2,
    #            self.points[1][0]*self.points[1][1],self.points[1][1]**2],
    #           [1, self.points[2][0], self.points[2][1],self.points[2][0]**2,
    #            self.points[2][0]*self.points[2][1],self.points[2][1]**2],
    #           [1, self.points[3][0], self.points[3][1],self.points[3][0]**2,
    #            self.points[3][0]*self.points[3][1],self.points[3][1]**2],
    #           [1, self.points[4][0], self.points[4][1],self.points[4][0]**2,
    #            self.points[4][0]*self.points[4][1],self.points[4][1]**2],
    #           [1, self.points[5][0], self.points[5][1],self.points[5][0]**2,
    #            self.points[5][0]*self.points[5][1],self.points[5][1]**2]]

    #     try :
    #         inv_a = a.inverse()
    #     except ValueError:
    #         self.plot()
    #         print(self._area())
    #         raise FlatElementError('form function bug')
    #     x1 = inv_a.dot([1,0,0,0,0,0])
    #     x2 = inv_a.dot([1,0,0,0,0,0])
    #     x3 = inv_a.dot([1,0,0,0,0,0])
    #     x4=inv_a.dot([1,0,0,0,0,0])

    #     return x1, x2, x3

    def _area(self):
        u_vect = self.points[1] - self.points[0]
        v_vect = self.points[2] - self.points[0]
        return abs(u_vect.cross(v_vect)) / 2

    # def point_belongs(self, point):
    #     polygon = volmdlr.wires.ClosedPolygon2D(self.points)
    #     point_belongs = polygon.point_belongs(point)
    #     return point_belongs

    # def rotation(self, center, angle, copy=True):
    #     if copy:
    #         return TriangularElement([pt.rotation(center, angle, copy=True)
    #                                   for pt in self.points])
    #     else:
    #         for pt in self.points:
    #             pt.Rotation(center, angle, copy=False)

    # def translation(self, offset, copy=True):
    #     if copy:
    #         return TriangularElement([pt.translation(offset, copy=True)
    #                                   for pt in self.points])
    #     else:
    #         for pt in self.points:
    #             pt.translation(offset, copy=False)

    def axial_symmetry(self, line):
        new_points = []
        for point in self.points:
            new_points.append(point.axial_symmetry(line))
        return self.__class__(new_points)

    # def axial_symmetry(self, line, copy=True):
    #     p1, p2 = line.points
    #     symmetric_points = []
    #     for point in self.points:
    #         u = p2 - p1
    #         t = (point-p1).dot(u) / u.norm()**2
    #         projection = p1 + t * u
    #         symmetric_point = volmdlr.Point2D(*(2 * projection - point))
    #         symmetric_points.append(symmetric_point)
    #     if copy:
    #         return TriangularElement(symmetric_points)
    #     else:
    #         for i, point in enumerate(self.points):
    #             point = symmetric_points[i]

    # def triangle_to_polygon(self):
    #     points = self.points
    #     return volmdlr.wires.ClosedPolygon2D(points)


class TriangularElement2D(TriangularElement, vmw.ClosedPolygon2D):
    """ Class to define a 2D triangular element. """
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, points, name: str = ''):
        super().__init__(points)
        # self.points = points
        self.name = name
        self.linear_elements = self._to_linear_elements()
        self.form_functions = self._form_functions()
        # self._line_segments = None

        # self.line_segments = self.get_line_segments()
        self.center = (self.points[0] + self.points[1] + self.points[2]) / 3

        self.area = self._area()
        # vmw.Triangle.__init__(self, points)

    def _to_linear_elements(self):
        vec1 = vm.Vector2D(self.points[1].x - self.points[0].x,
                           self.points[1].y - self.points[0].y)
        vec2 = vm.Vector2D(self.points[2].x - self.points[1].x,
                           self.points[2].y - self.points[1].y)
        vec3 = vm.Vector2D(self.points[0].x - self.points[2].x,
                           self.points[0].y - self.points[2].y)
        normal1 = vm.Vector2D(-vec1.y, vec1.x)
        normal2 = vm.Vector2D(-vec2.y, vec2.x)
        normal3 = vm.Vector2D(-vec3.y, vec3.x)
        normal1 = normal1.unit_vector()
        normal2 = normal2.unit_vector()
        normal3 = normal3.unit_vector()
        if normal1.dot(vec2) < 0:
            normal1 = - normal1
        if normal2.dot(vec3) < 0:
            normal2 = - normal2
        if normal3.dot(vec1) < 0:
            normal3 = - normal3
        linear_element_1 = LinearElement(self.points[0], self.points[1],
                                         normal1)
        linear_element_2 = LinearElement(self.points[1], self.points[2],
                                         normal2)
        linear_element_3 = LinearElement(self.points[2], self.points[0],
                                         normal3)
        return [linear_element_1, linear_element_2, linear_element_3]

    def _form_functions(self):
        a_matrix = vm.Matrix33(1, self.points[0].x, self.points[0].y,
                               1, self.points[1].x, self.points[1].y,
                               1, self.points[2].x, self.points[2].y)
        try:
            inv_a = a_matrix.inverse()
        except ValueError as expt:
            self.plot()
            print('buggy element area', self.area)
            raise FlatElementError('form function bug') from expt
        x_1 = inv_a.vector_multiplication(vm.X3D)
        x_2 = inv_a.vector_multiplication(vm.Y3D)
        x_3 = inv_a.vector_multiplication(vm.Z3D)
        return x_1, x_2, x_3

    def _area(self):
        u_vect = self.points[1] - self.points[0]
        v_vect = self.points[2] - self.points[0]
        return abs(u_vect.cross(v_vect)) / 2

    def axial_symmetry(self, line):
        new_points = []
        for point in self.points:
            new_points.append(point.axial_symmetry(line))
        return self.__class__(new_points)

    def plot(self, ax=None, edge_style: EdgeStyle = EdgeStyle(), point_numbering=False,
             fill=False, fill_color='w'):
        if ax is None:
            _, ax = plt.subplots()
            ax.set_aspect('equal')

        if fill:
            x = [p[0] for p in self.points]
            y = [p[1] for p in self.points]
            plt.fill(x, y, facecolor=fill_color, edgecolor="k")
            return ax
        if point_numbering:
            for index_point, point in enumerate(self.points):
                ax.text(*point, f'point {index_point + 1}', ha='center', va='top')
        for p1, p2 in zip(self.points, self.points[1:] + [self.points[0]]):
            if edge_style.width is None:
                edge_style.width = 1
            if edge_style.plot_points:
                ax.plot([p1.x, p2.x], [p1.y, p2.y], color=edge_style.color,
                        marker='o', linewidth=edge_style.linewidth)
            else:
                ax.plot([p1.x, p2.x], [p1.y, p2.y], color=edge_style.color,
                        linewidth=edge_style.linewidth)
        return ax


class QuadrilateralElement2D(vmw.ClosedPolygon2D):
    """ Class to define a 2D quadrilateral element. """

    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, points: List[volmdlr.Point2D]):
        self.points = points
        # self.linear_elements = self._to_linear_elements()
        # self.form_functions = self._form_functions()
        # self.line_segments = self.line_segments
        self.center = self.center_of_mass()

        self.area = self.area()
        self._line_segments = None
        vmw.ClosedPolygon2D.__init__(self, points)


class TriangularElement3D(TriangularElement, vmw.ClosedPolygon3D):
    """ Class to define a 3D triangular element. """

    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, points):
        self.points = points
        # self.linear_elements = self._to_linear_elements()
        # self.form_functions = self._form_functions()
        # self.line_segments = self.line_segments
        self.center = (self.points[0] + self.points[1] + self.points[2]) / 3

        # self.area = self._area()
        self._line_segments = None
        super().__init__(points)

    def axial_symmetry(self, line):
        """ Returns a symmetric new element with respect to the given line. """
        new_points = []
        for point in self.points:
            new_points.append(point.axial_symmetry(line))
        return self.__class__(new_points)


class TetrahedralElement(DessiaObject):
    """ Class to define a 3D tetrahedral element. """

    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, points, name: str = ''):
        self.points = points
        self.name = name
        # self.linear_elements = self._to_linear_elements()
        self.form_functions = self._form_functions()
        # self.line_segments = self._line_segments()
        self.center = (self.points[0] + self.points[1] + self.points[2] + self.points[3]) / 4
        self.triangular_elements = self._triangular_elements()

        self.volume = self._volume()
        DessiaObject.__init__(self, name=name)

    def _triangular_elements(self):

        indices_combinations = list(combinations(list(range(len(self.points))), r=3))
        # indices_combinations = [x for x in combinations(list(range(len(self.points))), r=3)]
        triangular_elements = []

        for indices in indices_combinations:
            triangular_elements.append(TriangularElement3D([self.points[indices[0]],
                                                            self.points[indices[1]],
                                                            self.points[indices[2]]]))

        return triangular_elements

    def plot(self, ax=None, edge_style: EdgeStyle = EdgeStyle()):
        if ax is None:
            ax = plt.figure().add_subplot(projection='3d')
        for point in self.points:
            point.plot(ax=ax)
        for triangle in self.triangular_elements:
            triangle.plot(ax=ax, edge_style=edge_style)
        return ax

    def _volume(self):

        data = []
        for i in range(3):
            data.extend([*self.points[i + 1] - self.points[0]])

        return abs(1 / 6 * (npy.linalg.det(npy.array(data).reshape(3, 3))))

    def _form_functions(self):
        coeff = [1, -1, 1, -1]
        form_funct = []
        for i in range(4):
            data_alpha, data_gamma, data_betha, data_delta = [], [], [], []
            for c_coef in range(4):
                if c_coef != i:
                    data_alpha.extend([self.points[c_coef].x, self.points[c_coef].y, self.points[c_coef].z])
                    data_gamma.extend([1, self.points[c_coef].x, self.points[c_coef].z])
                    data_betha.extend([1, self.points[c_coef].y, self.points[c_coef].z])
                    data_delta.extend([1, self.points[c_coef].x, self.points[c_coef].y])

            form_funct.append([(coeff[i] * (npy.linalg.det(npy.array(data_alpha).reshape(3, 3)))),
                               ((-1) * coeff[i] * (npy.linalg.det(npy.array(data_betha).reshape(3, 3)))),
                               (coeff[i] * (npy.linalg.det(npy.array(data_gamma).reshape(3, 3)))),
                               ((-1) * coeff[i] * (npy.linalg.det(npy.array(data_delta).reshape(3, 3))))])

        return form_funct[0], form_funct[1], form_funct[2], form_funct[3]


class TetrahedralElementQuadratic(DessiaObject):
    """ Class to define a 3D quadratic tetrahedral element. """

    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, points, name: str = ''):
        self.points = points
        self.name = name
        DessiaObject.__init__(self, name=name)


class ElementsGroup(DessiaObject):
    """Defines a group of elements."""

    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, elements, name: str):
        self.elements = elements
        self.nodes = self._nodes()
        self.name = name

        self._elements_per_node = None

        DessiaObject.__init__(self, name=name)

    def _nodes(self):
        nodes = set()
        for element in self.elements:
            for point in element.points:
                nodes.add(point)
        return nodes

    def point_to_element(self, point):
        """
        Checks if a point belongs to an element.

        :param point: A point2d/3d
        :type point: Point2D/Point3D

        :return: The input Element OR None
        """

        for element in self.elements:
            if element.point_belongs(point):
                return element
        return None

    @property
    def elements_per_node(self):
        """Maps the elements of each node."""
        if self._elements_per_node is not None:
            return self._elements_per_node

        dict_node_element = {}
        for element in self.elements:
            for point in element.points:
                try:
                    dict_node_element[point].append(element)
                except KeyError:
                    dict_node_element[point] = [element]
        self._elements_per_node = dict_node_element
        return dict_node_element

    # def rotation(self, center, angle, copy=True):
    #     if copy:
    #         return Mesh([elem.rotation(center, angle, copy=True)
    #                       for elem in self.elements])
    #     else:
    #         for elem in self.elements:
    #             elem.rotation(center, angle, copy=False)

    # def translation(self, offset, copy=True):
    #     if copy:
    #         return Mesh([elem.translation(offset, copy=True)
    #                       for elem in self.elements])
    #     else:
    #         for elem in self.elements:
    #             elem.translation(offset, copy=False)

    def plot(self, ax=None, color='k'):  # , fill=False):
        """
        Plot an ElementGroup.
        """

        if ax is None:
            _, ax = plt.subplots()
            ax.set_aspect('equal')
        for element in self.elements:
            element.plot(ax=ax, edge_style=EdgeStyle(color=color))  # fill=fill
        return ax


class Mesh(DessiaObject):
    """Defines a mesh."""

    _standalone_in_db = True
    _non_serializable_attributes = ['node_to_index']
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, elements_groups: List[ElementsGroup]):
        self.elements_groups = elements_groups
        self.nodes = self._set_nodes_number()
        self.node_to_index = {self.nodes[i]: i for i in range(len(self.nodes))}
        self._nodes_correction = {}
        self._gmsh = None
        DessiaObject.__init__(self, name='')

    def _set_nodes_number(self):
        nodes = set()
        for elements_group in self.elements_groups:
            for element in elements_group.elements:
                for point in element.points:
                    nodes.add(point)
                # nodes.add(element.points[0])
                # nodes.add(element.points[1])
                # nodes.add(element.points[2])
        return tuple(nodes)

    def point_to_element(self, point):
        for element_group in self.elements_groups:
            element = element_group.point_to_element(point)
            if element is not None:
                return element
        return None

    def plot(self, ax=None):
        if ax is None:
            if self.elements_groups[0].elements[0].__class__.__name__[-2::] == '2D':
                _, ax = plt.subplots()
                ax.set_aspect('equal')
            else:
                ax = plt.figure().add_subplot(projection='3d')
        for elements_group in self.elements_groups:
            elements_group.plot(ax=ax)
        return ax

    def bounding_rectangle(self):
        nodes = self.nodes
        x, y = [], []
        for n in nodes:
            x.append(n.x)
            y.append(n.y)
        if len([*nodes[0]]) == 2:
            return min(x), max(x), min(y), max(y)

        z = [n.z for n in nodes]
        return min(x), max(x), min(y), max(y), min(z), max(z)

    # def delete_duplicated_nodes(self, tol=1e-4):
    #     mesh = self.__class__(self.elements_groups[:])
    #     nodes_list = list(mesh.nodes[:])
    #     nodes_index = []

    #     for i, node in enumerate(nodes_list):
    #         for j in range(i + 1, len(nodes_list)):
    #             dist = node.point_distance(nodes_list[j])
    #             if dist < tol:
    #                 nodes_index.append((j, i))

    #     if nodes_index:
    #         nodes_index = sorted(nodes_index, key=lambda item: item[0], reverse=True)
    #         for _, index in enumerate(nodes_index):
    #             nodes_list.pop(index[0])
    #             for group in mesh.elements_groups:
    #                 if mesh.nodes[index[0]] in group.nodes:
    #                     dict_node_element = group.elements_per_node
    #                     for element in dict_node_element[mesh.nodes[index[0]]]:
    #                         element.points[element.points.index(mesh.nodes[index[0]])] = mesh.nodes[index[1]]

    #         mesh.nodes = nodes_list
    #         mesh.node_to_index = {mesh.nodes[i]: i for i in range(len(mesh.nodes))}

    #     return mesh

    def nodes_correction(self, reference_index, tol=1e-4):
        if not self._nodes_correction:
            nodes_reference = self.elements_groups[reference_index].nodes
            groups = self.elements_groups[:]
            groups.pop(reference_index)
            nodes_correction = {}

            for group in groups:
                for node in group.nodes:
                    for node_ref in nodes_reference:
                        d = node.point_distance(node_ref)
                        if 1e-8 < d < 1e-4:
                            nodes_correction[node] = node_ref

            self._nodes_correction = nodes_correction

        return self._nodes_correction

    def delete_duplicated_nodes(self, reference_index, tol=1e-4):
        """Delete duplicated nodes."""
        groups = self.elements_groups[:]
        groups.pop(reference_index)

        nodes_correction = self.nodes_correction(reference_index, tol)

        count = 0
        old_elements, new_elements = set(), set()
        for index_g, group in enumerate(groups):
            elements = []
            for element in group.elements:
                new = False
                points = []

                for point in element.points:
                    correc_point = nodes_correction.get(point)
                    if correc_point is not None:
                        points.append(correc_point)
                        count += 1
                        old_elements.add(element)
                        new = True
                    else:
                        points.append(point)

                elements.append(element.__class__(points))
                if new:
                    new_elements.add(element.__class__(points))

            groups[index_g] = group.__class__(elements, name='')

        groups.insert(reference_index, self.elements_groups[reference_index])

        return self._helper_create_mesh(groups)

    def _helper_create_mesh(self, groups):
        mesh = self.__class__(groups)
        mesh.gmsh = self.gmsh
        mesh.set_nodes_correction(self.get_nodes_correction())
        return mesh

    def get_nodes_correction(self):
        """
        A getter method for nodes_correction private variable.

        :return: A dict of nodes_correction
        :rtype: dict
        """

        return self._nodes_correction

    def set_nodes_correction(self, nodes_correction):
        """
        A setter method for nodes_correction private variable.

        :param nodes_correction: A dict of nodes_correction
        :type nodes_correction: dict
        """

        if not isinstance(nodes_correction, dict):
            raise ValueError("It must be volmdlr.GmshParser class")
        self._nodes_correction = nodes_correction

    @property
    def gmsh(self):
        """
        A property to get gmsh (a private variable).

        :return: A gmsh_parser data
        :rtype: GmshParser
        """

        return self._gmsh

    @gmsh.setter
    def gmsh(self, gmsh_parser):
        """
        A setter for gmsh (a private variable).

        :param gmsh_parser: A gmsh_parser data
        :type gmsh_parser: GmshParser
        """

        if not isinstance(gmsh_parser, volmdlr.gmsh_vm.GmshParser):
            raise ValueError("It must be volmdlr.GmshParser class")
        self._gmsh = gmsh_parser

    def copy(self):
        mesh = self.__class__(elements_groups=self.elements_groups[:])
        mesh.set_nodes_correction(self.get_nodes_correction())
        mesh.gmsh = self.gmsh

        return mesh



"""
volmdlr cad simplification module.
"""
from dessia_common.core import DessiaObject
import volmdlr
from volmdlr import shells, primitives3d, wires
from volmdlr.cloud import PointCloud3D


class OctreeBlockSimplify(DessiaObject):
    """
    Initializes an instance of OctreeBlockSimplify.

    :param closed_shell: A closed shell object representing the geometry to simplify.
    :param name: Optional. A name for the OctreeBlockSimplify instance.

    """

    def __init__(self, closed_shell: shells.ClosedShell3D, name: str = ''):
        self.closed_shell = closed_shell
        self.block = primitives3d.Block.from_bounding_box(self.closed_shell.bounding_box)
        DessiaObject.__init__(self, name=name)

    def get_outside_blocks(self, deepness: int = 3):
        """
        Gets octree blocks outside the given closed shell.

        :param deepness: Optional. The depth of octree subdivision (default: 3).

        :return: A list of octree blocks that are outside the given closed shell.

        """
        block_volume = self.block.volume()
        min_block_volume = block_volume / (8 ** deepness)
        dict_faces_to_box = {box: [] for box in self.block.octree()}
        for box, list_faces in dict_faces_to_box.items():
            for face in self.closed_shell.faces:
                if (
                        box.is_face_intersecting(face)
                        or box.is_face_inside(face)
                        or box.bounding_box.distance_to_bbox(face.bounding_box) <= 1e-6
                ):
                    list_faces.append(face)
        not_intersecting_boxes = []
        inside_boxes = []

        for octant in self.block.octree():
            dict_faces_to_box_ = {octant: dict_faces_to_box[octant]}
            octant_not_intersecting_boxes = []

            while dict_faces_to_box_:
                box, list_faces = next(iter(dict_faces_to_box_.items()))
                if not list_faces:
                    if not box.is_inside_shell(self.closed_shell):
                        octant_not_intersecting_boxes.append(box)
                    else:
                        inside_boxes.append(box)
                    del dict_faces_to_box_[box]
                    continue

                divided_boxes = box.octree()
                if divided_boxes[0].volume() >= min_block_volume:
                    for divided_box in divided_boxes:
                        new_list_faces = [
                            face
                            for face in list_faces
                            if (
                                    divided_box.is_face_intersecting(face)
                                    or divided_box.is_face_inside(face)
                                    or divided_box.bounding_box.distance_to_bbox(face.bounding_box) <= 1e-6
                            )
                        ]
                        dict_faces_to_box_[divided_box] = new_list_faces

                del dict_faces_to_box_[box]

            not_intersecting_boxes.append(octant_not_intersecting_boxes)

        return not_intersecting_boxes

    def simplify(self, precision: int):
        """
        Simplify the given closed shell.

        :param precision: The precision level of simplification.

        :return: The simplified closed shell.

        """
        outside_blocks = self.get_outside_blocks(deepness=precision)
        closed_shells = []
        for octant in outside_blocks:
            if not octant:
                continue
            list_shells = shells.union_list_of_shells(octant)
            closed_shells.extend(list_shells)
        closed_shells_ = shells.union_list_of_shells(closed_shells.copy())
        closed_shells_ = sorted(closed_shells_, key=lambda shell: shell.volume(), reverse=True)
        simplified_closed_shell = self.block.subtract_to_closed_shell(closed_shells_[0])[0]
        return simplified_closed_shell


class TrippleExtrusionSimplify(DessiaObject):
    """
    Initialize an instance of TrippleExtrusionSimplify.

    :param volume_model: The volume model to simplify.
    :param name (str): Optional. A name for the TrippleExtrusionSimplify instance.

    """

    def __init__(self, volume_model, name: str = ''):
        self.volume_model = volume_model
        DessiaObject.__init__(self, name=name)

    def simplify(self):
        """
        Simplify the volume model using triple extrusion simplification.

        :return: The simplified volume model.

        """
        points = []
        for primitive in self.volume_model.primitives:
            tri = primitive.triangulation()
            points.extend(tri.points)
        cloud = PointCloud3D(points)

        model = self.extrusion_union_cloud_simplifier(cloud)
        return model

    @staticmethod
    def extrusion_union_cloud_simplifier(cloud3d):
        """
        Simplify a point cloud using extrusion and union operations.

        :param cloud3d: The 3D point cloud to simplify.

        :return: The simplified shell.

        """
        simplified_shell = None
        list_shells = []

        bbox = cloud3d.bounding_box
        lx = bbox.xmax - bbox.xmin
        ly = bbox.ymax - bbox.ymin
        lz = bbox.zmax - bbox.zmin
        for w_vector, length, center in zip([volmdlr.X3D, volmdlr.Y3D, volmdlr.Z3D], [lx, ly, lz], bbox.center):
            u_vector = w_vector.random_unit_normal_vector()
            v_vector = w_vector.cross(u_vector)
            cloud2d = cloud3d.to_2d(volmdlr.O3D, u_vector, v_vector)
            polygon2d = cloud2d.to_polygon(convex=True)
            contour2d = wires.Contour2D(polygon2d.line_segments)
            frame = volmdlr.Frame3D(((center - 0.5 * length) * w_vector).to_point(), u_vector, v_vector, w_vector)
            dir_shell = primitives3d.ExtrudedProfile(frame, contour2d, [], length)
            dir_shell.merge_faces()

            list_shells.append(dir_shell)
            if simplified_shell is None:
                simplified_shell = dir_shell
            else:
                list_shells = simplified_shell.intersection(dir_shell)
                simplified_shell = list_shells[0]

        return simplified_shell
